"""Convert Epubs from the Ghbook / Ghaemiyeh (Qaimiyya) library \
to OpenITI mARkdown.

The converter has two main functions:
* convert_file: convert a single epub file.
* convert_files_in_folder: convert all epub files in a given folder

Usage examples:
    >>> folder = "test/ghaemiyeh/epub/"
    >>> meta_fp = "test/ghaemiyeh/all_books/meta/all_metadata.json"
    >>> from epub_converter_ghaemiyeh import convert_file, convert_files_in_folder
    >>> src_fp = folder+"000008.epub"
    >>> convert_file(src_fp, meta_fp, dest_fp=folder+"converted/ghaemiyeh000008")
    >>> convert_files_in_folder(folder, meta_fp, dest_folder=folder+"converted")
    Converting all files in folder test/ghaemiyeh/epub with extensions ['epub']

Both functions use the ghaemiyehEpubConverter class to do the heavy lifting.
The ghaemiyehEpubConverter is a subclass of the GenericEpubConverter,
which in turn is a subclass of the GenericConverter
from the generic_converter module:

GenericConverter
    \_ GenericEpubConverter
            \_ ghaemiyehEpubConverter

Methods of both classes:

(methods of GenericConverter are inherited by GenericEpubConverter;
methods of GenericConverter with the same name
in GenericEpubConverter are overwritten by the latter)

=========================== ========================= =======================
generic_converter           epub_converter_generic    epub_converter_ghaemiyeh 
=========================== ========================= =======================
__init__                    __init__                  __init__ 
convert_files_in_folder     (inherited)               (inherited)
convert file                (inherited)               (inherited)
make_dest_fp                (inherited - generic!)    (inherited - generic!)
get_metadata                (inherited - generic!)    get_metadata
get_data                    get_data                  (inherited)
pre_process                 (inherited)               (inherited)
add_page_numbers            (inherited - generic!)    (inherited - generic!)
add_structural_annotations  (inherited - generic!)    (inherited - generic!) 
remove_notes                remove_notes              (inherited)
reflow                      (inherited)               (inherited)
add_milestones              (inherited)               (inherited)
post_process                (inherited - generic!)    post_process
compose                     (inherited)               (inherited)
save_file                   (inherited)               (inherited)
                            convert_html2md           convert_html2md
                            inspect_epub              (inherited)
                            sort_html_files_by_toc    sort_html_files_by_toc
                            add_unique_tags           (inherited)
=========================== ========================= =======================


Examples:
    >>> from epub_converter_ghaemiyeh import ghaemiyehEpubConverter
    >>> from helper.yml2json import yml2json
    >>> folder = "test/"
    >>> fn = "26362727.epub"
    >>> hc = ghaemiyehEpubConverter(dest_folder="test/converted")
    >>> hc.VERBOSE = False
    >>> meta_fp = "ghaemiyeh/all_books/meta/all_metadata.json"
    >>> hc.metadata_file = meta_fp
    >>> hc.convert_file(folder+fn)

    #>>> hc.convert_files_in_folder(folder)

"""

import os
import json
import shutil
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.epub_converter_generic import GenericEpubConverter
from openiti.new_books.convert.helper import html2md_ghaemiyeh
from openiti.new_books.convert.helper.yml2json import yml2json


def load_metadata(meta_fp):
    with open(meta_fp, mode="r", encoding="utf-8") as file:
        meta = file.read().splitlines()
    header = meta.pop(0).split("\t")
    id_col = header.index("book_id")
    d = dict()
    for row in meta:
        try:
            row = row.split("\t")
            book_id = int(row[id_col])
        except:
            print("no book_id found!")
            continue
        d[book_id] = dict()
        for i, key in enumerate(header):
            d[book_id][key] = row[i]
            
    return d
    

def convert_file(fp, meta_fp, dest_fp=None, verbose=False, overwrite=False):
    """Convert one file to OpenITI format.

    Args:
        fp (str): path to the file that must be converted.
        meta_fp (str): path to the yml file containing the ghaemiyeh metadata
        dest_fp (str): path to the converted file.

    Returns:
        None
    """
    conv = ghaemiyehEpubConverter(overwrite=overwrite)
    conv.VERBOSE = verbose
    conv.metadata_file = meta_fp
    conv.metadata_dic = load_metadata(meta_fp)
    conv.convert_file(fp, dest_fp=dest_fp)

##def convert_multifile_text(folder, meta_fp, dest_folder, verbose=False):
##    for i, fn in enumerate(os.listdir(folder)):
##        if i == 0:
##            dest_fp = os.path.join(dest_folder, os.path.splitext(fn)[0])
##        
        

def convert_files_in_folder(src_folder, meta_fp, dest_folder=None, verbose=False,
                            extensions=["epub"], exclude_extensions=["yml"],
                            fn_regex=None, overwrite=False):
    """Convert all files in a folder to OpenITI format.\
    Use the `extensions` and `exclude_extensions` lists to filter\
    the files to be converted.

    Args:
        src_folder (str): path to the folder that contains
            the files that must be converted.
        meta_fp (str): path to the yml file containing the ghaemiyeh metadata
        dest_folder (str): path to the folder where converted files
            will be stored.
        extensions (list): list of extensions; if this list is not empty,
            only files with an extension in the list should be converted.
        exclude_extensions (list): list of extensions;
            if this list is not empty,
            only files whose extension is not in the list will be converted.
        fn_regex (str): regular expression defining the filename pattern
            e.g., "-(ara|per)\d". If `fn_regex` is defined,
            only files whose filename matches the pattern will be converted.

    Returns:
        None
    """
    msg = "Converting all files in folder {} with extensions {}"
    print(msg.format(src_folder, extensions))
    conv = ghaemiyehEpubConverter(overwrite=overwrite)
    conv.VERBOSE = verbose
    conv.metadata_dic = load_metadata(meta_fp)
    conv.metadata_file = meta_fp
    conv.convert_files_in_folder(src_folder, dest_folder=dest_folder,
                                 extensions=extensions,
                                 exclude_extensions=exclude_extensions,
                                 fn_regex=fn_regex)


################################################################################




class ghaemiyehEpubConverter(GenericEpubConverter):
    def __init__(self, dest_folder=None, overwrite=True):
        super().__init__(dest_folder=dest_folder, overwrite=overwrite)
        self.toc_fn = "content.opf"
        self.metadata_file = None


    def convert_files_in_folder(self, source_folder, dest_folder=None,
                                extensions=[], exclude_extensions=[],
                                fn_regex=None):
        """Convert all files in a folder to OpenITI format.\
        Use the `extensions` and `exclude_extensions` lists to filter\
        the files to be converted.

        Args:
            source_folder (str): path to the folder that contains
                the files that must be converted.
            extensions (list): list of extensions; if this list is not empty,
                only files with an extension in the list should be converted.
            exclude_extensions (list): list of extensions;
                if this list is not empty,
                only files whose extension is not in the list will be converted.
            fn_regex (str): regular expression defining the filename pattern
                e.g., "-(ara|per)\d". If `fn_regex` is defined,
                only files whose filename matches the pattern will be converted.

        Returns:
            None
        """
        failed = []
        if dest_folder:
            self.dest_folder = dest_folder
        fp_list = self.filter_files_in_folder(source_folder, extensions,
                                              exclude_extensions, fn_regex)
        for fp in fp_list:
            #self.inspect_epub(fp)
            #print(fp)
            try:
                self.convert_file(fp)
                #print("converted!")
            except Exception as e:
                print(fp)
                print("  >> ERROR:", e)
                failed.append((fp, e))
                

        # deal with multivolume texts:
        print("checking multivolume texts")
        all_book_ids = [os.path.splitext(os.path.split(fp)[-1])[0] for fp in fp_list]
        multivol_book_ids = dict()
        for book_id in all_book_ids:
            if len(book_id) > 7:
                main_id = re.sub("^10+(\d+)\d\d\d$", r"\1", book_id)
                if not main_id in multivol_book_ids:
                    multivol_book_ids[main_id] = []
                multivol_book_ids[main_id].append(book_id)
        for main_id in multivol_book_ids:
            combined = ""
            endnotes = []
            outfp = os.path.join(self.dest_folder, main_id + ".automARkdown")
            if os.path.exists(outfp):  # if a file exists containing all volumes, don't join separate volumes
                continue
            outfp = re.sub("\.automARkdown", "Vols.automARkdown", outfp)
            if os.path.exists(outfp):  # if a file exists containing all volumes, don't join separate volumes
                continue
            print("joining", main_id)
            for i, book_id in enumerate(sorted(multivol_book_ids[main_id])):
                fp = os.path.join(self.dest_folder, book_id + ".automARkdown")
                try:
                    with open(fp, mode="r", encoding="utf-8") as file:
                        text = file.read()
                except:
                    print("file does not exist:", fp)
                    continue
                if i != 0:
                    text = re.split("#META#Header#End#", text)[-1]
                    text = re.sub("PageV01", "PageV{0:02d}".format(i+1), text)
                if re.findall("### \|EDITOR\|[ \r\n]+ENDNOTES:?", text):
                    text, notes = re.split("### \|EDITOR\|[ \r\n]+ENDNOTES:?", text)
                    endnotes.append(notes)
                combined += text
                
            if endnotes:
                combined += "\n\n### |EDITOR|\nENDNOTES:\n\n" + "\n\n".join(endnotes)
            
            if combined:
                with open(outfp, mode="w", encoding="utf-8") as file:
                    file.write(combined)
            else:
                print("No contents for", outfp)


        print("Converting all files done")
        if failed:
            print("These files failed to convert:")
            for fp, e in failed:
                print(fp, e)
        
                

    def sort_html_files_by_toc(self, zp, toc_fp, html_files):
        """Gets the table of contents from the Epub file.

        Args:
            zp: zipfile object
            toc_fp (str): filepath to the table of contents of the epub.
            html_files(list): an unordered list of the html files in the epub.

        Returns:
            toc (list): a list of filepaths to the html files
                in the epub file, in the order specified by the
                table of contents
        """
        html_files_dict = {os.path.split(fp)[-1] : fp for fp in html_files}
        toc_data = zp.read(toc_fp)
        toc_data = codecs.decode(toc_data, "utf-8")
        soup = BeautifulSoup(toc_data)
        toc_ol = soup.find("spine")
        toc = []
        for item in toc_ol.find_all("itemref"):
            fn = os.path.split(item.get("idref"))[-1]
            if fn in html_files_dict:
                toc.append(html_files_dict[fn])
        return toc

    def convert_html2md(self, html):
        """Use custom html to mARKdown function for ghaemiyeh epubs."""
        text = html2md_ghaemiyeh.markdownify(html)
        return text

    def remove_notes(self, text):
        """Remove the footnotes from the text.

        Args:
            text (str)

        Returns:
            text_without_notes (str): text from which the notes were removed.
            notes (str): the footnotes converted to endnotes
                (if there are endnotes: add the endnote_splitter before them)
        """
        print("REMOVING NOTES")
        text_without_notes = ""
        notes = ""
        spl = re.split("(FOOTNOTE.*\n+)", text)
        spl = [x for x in spl if x != ""]
        i = 0
        #last_text_fragm = 0
        for t in spl:
            if not t.strip().startswith("FOOTNOTE"):
                text_without_notes += t
                #last_text_fragm = i
            else:
                fn_text = t.strip()[8:]
                if spl[i-1].strip().startswith("FOOTNOTE"):
                    notes += "{}\n\n".format(fn_text)
                else:
                    #p = re.findall("PageV\d+P\d+", spl[last_text_fragm])
                    #p = re.findall("PageV\d+P\d+", spl[i-1])
                    p = re.findall("# ص *: *(\d+)", spl[i-1])
                    if p:
                        notes += "PageV01P{:03d}:\n\n{}\n\n".format(int(p[-1]), fn_text)
                    else:
                        section_rgx = "(### [\|$]+) (.*)"
                        #sections = re.findall(section_rgx, spl[last_text_fragm])
                        sections = re.findall(section_rgx, spl[i-1])
                        if sections:
                            msg = "Notes on section ({})"
                            msg +=  "(no page numbers):\n\n{}\n\n"
                            notes += msg.format(sections[0][1], fn_text)
                        else:
                            notes += "(no page number):\n\n{}\n\n".format(fn_text)
            i += 1
        if notes:
            notes = self.endnote_splitter + notes
        return text_without_notes, notes

    def get_metadata(self, metadata_fp):
        """Custom method to get the metadata of the ghaemiyeh epub file."""
        source_fp = self.source_fp
        bookID = os.path.split(source_fp)[1]
        bookID = int(os.path.splitext(bookID)[0])
        meta_dic = self.metadata_dic[bookID]
        meta = ["#META# {}: {}".format(k,v) for k,v in sorted(meta_dic.items())]
        return self.magic_value + "\n".join(meta) + self.header_splitter

    def post_process(self, text):
        """Custom post-processing for ghaemiyeh texts"""
        
        text = re.sub("# ص *: *(\d+)", r"PageV01P\1", text)
        text = re.sub("#(?! \d).+ج *(\d+)[، ]*ص *: *(\d+)", r"PageV\1P\2", text)
        text = re.sub("PageV(\d)P", r"PageV0\1P", text)
        text = re.sub("PageV01P(\d)(?!\d)", r"PageV01P00\1", text)
        text = re.sub("PageV01P(\d\d)(?!\d)", r"PageV01P0\1", text)
        processed = super().post_process(text)
        return processed
        


if __name__== "__main__":
    #import doctest
    #doctest.testmod()
    #input("Testing finished. Continue?")

    # identify the location of the yml file containing the metadata:
    meta_fp = r"H:\OpenITI\new_books\Ghbook\books_AR_FA_UR_cleaned.tsv"
    src_folder = r"H:\OpenITI\new_books\Ghbook\epub\AR"
    dest_folder = r"H:\OpenITI\new_books\Ghbook\converted"
    convert_files_in_folder(src_folder, meta_fp, dest_folder=dest_folder,
                            verbose=False, overwrite=False)
##    hc.metadata = yml2json(meta_fp, container={})
    
##    fp = r"test\26362727.epub"
##    hc.convert_file(fp)
##    print("converted ghaemiyeh epub", fp)
##
##    hc.convert_files_in_folder("test/ghaemiyeh")
##    print("converted all epub files in folder", "test/ghaemiyeh")




"""Convert TEI xml to Markdown.

This program contains a sub-class of html2md.MarkdownConverter,
which in turn is an adaptation of python-markdownify
(https://github.com/matthewwithanm/python-markdownify)
to output OpenITI mARkdown.

IMPORTANT: since TEI indicates page beginning (<pb/>)
and OpenITI mARkdown page numbers are at the bottom of a page,
page numbers should be pre-processed in a text
before feeding it to the markdownify function.
The tei2md.preprocess_page_numbers function can be used for this. 


You can use the tei2md.TeiConverter class as a base class
and subclass it to add methods, adapt the post-processing method etc.

E.g.::
    def Class GRAR_converter(tei2md.TeiConverter):   
        def post_process_md(self, text):
            text = super().post_process_md(text)

            # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag

            text = re.sub("\\\\n+DELETE_PREVIOUS_BLANKLINES", "", text)

            # replace placeholders for spaces in tables: 

            text = re.sub("ç", " ", text)
            return text

This table shows the methods and classes the TeiConverter
inherits from html2md.MarkdownConverter,
which methods it overwrites,
and which methods it adds: 

=========================== ==========================
html2md.MarkdownConverter   tei2md.TeiConverter
=========================== ==========================
class DefaultOptions        (inherited)
class Options               (inherited)
__init__                    (inherited)
convert                     convert
process_tag                 (inherited)
process_text                (inherited)
fill_out_columns            (inherited)
post_process_md             (inherited)
__getattr__                 (inherited)
should_convert_tag          (inherited)
indent                      (inherited)
create_underline_line       (inherited)
underline                   (inherited)
convert_a                   (inherited)
convert_b                   (inherited)
convert_blockquote          (inherited)
convert_br                  (inherited)
convert_em                  (inherited)
convert_hn                  (inherited)
convert_i                   (inherited)
convert_img                 (inherited)
convert_list                (inherited)
convert_li                  (inherited)
convert_p                   (inherited)
convert_strong              (inherited)
convert_table               (inherited)
convert_tr                  (inherited)
                            convert_div1
                            convert_div2
                            convert_div3
                            convert_div
                            convert_head
                            convert_lg
                            find_heading_level (dummy)
=========================== ==========================


Examples (doctests):

    Specific tei-related tags: 

    >>> import tei2md
    >>> h = 'abc\
             <div1 type="book" n="0" name="Preface">def</div1>\
             ghi'
    >>> tei2md.markdownify(h)
    'abc\\n\\n### | [book 0: Preface]\\n\\ndef\\n\\nghi'

    >>> h = 'abc\
             <div2 type="section" n="1">def</div1>\
             ghi'
    >>> tei2md.markdownify(h)
    'abc\\n\\n### || [section 1]\\n\\ndef\\n\\nghi'

    >>> h = 'abc\
             <div3 type="Aphorism">def</div1>\
             ghi'
    >>> tei2md.markdownify(h)
    'abc\\n\\n### ||| [Aphorism]\\n\\ndef\\n\\nghi'

    Divs without type are stripped:
    
    >>> h = 'abc\
             <div>def</div>\
             ghi'
    >>> tei2md.markdownify(h)
    'abc def ghi'

    <head> tags are converted to level-3 mARkdown headers by default:

    >>> h = 'abc\
             <head>def</head>\
             ghi'
    >>> tei2md.markdownify(h)
    'abc\\n\\n### ||| def\\n\\nghi'

    >>> h = 'abc\
             <lb/>def\
             <lb/>ghi'
    >>> tei2md.markdownify(h)
    'abc\\n~~def\\n~~ghi'

    >>> h = '\
    abc\
    <lg>\
      <l>line1</l>\
      <l>line2</l>\
      <l>line3</l>\
      <l>line4</l>\
    </lg>\
    def'
    >>> tei2md.markdownify(h)
    ' abc\\n# line1\\n# line2\\n# line3\\n# line4\\n\\ndef'


    In addition to these TEI tags, the converter also inherits methods
    from html2md.MarkdownConverter that deal with more standard html tags: 

    Headings: h1

    >>> import tei2md
    >>> h = '<h1>abc</h1>'
    >>> tei2md.markdownify(h)
    '\\n\\n### | abc\\n\\n'

    NB: heading style is OpenITI mARkdown style by default,
        but can be set to other styles as well:

    >>> h = '<h1>abc</h1>'
    >>> tei2md.markdownify(h, md_style=UNDERLINED)
    '\\n\\nabc\\n===\\n\\n'

    >>> h = '<h1>abc</h1>'
    >>> tei2md.markdownify(h, md_style=ATX)
    '\\n\\n# abc\\n\\n'

    Paragraphs (<p>):
    
    >>> h = "<p>abc</p>"
    >>> tei2md.markdownify(h)
    '\\n\\n# abc\\n\\n'

    >>> h = "<p>abc</p>"
    >>> tei2md.markdownify(h, md_style=ATX)
    '\\n\\nabc\\n\\n'


    Divs without type are stripped:
    
    >>> h = 'abc\
             <div>def</div>\
             ghi'
    >>> tei2md.markdownify(h)
    'abc def ghi'



    Spans without class or with an unsupported class are stripped:
    
    >>> h = 'abc <span>def</span> ghi'
    >>> tei2md.markdownify(h)
    'abc def ghi'
    
    >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
    >>> tei2md.markdownify(h)
    'abc def ghi'


    Links: 

    >>> h = '<a href="a/b/c">abc</a>'
    >>> tei2md.markdownify(h)
    '[abc](a/b/c)'

    
    Unordered lists: 

    >>> h = '<ul><li>item1</li><li>item2</li></ul>'
    >>> tei2md.markdownify(h)
    '\\n* item1\\n* item2\\n\\n'
    
    Ordered lists:

    >>> h = '<ol><li>item1</li><li>item2</li></ol>'
    >>> tei2md.markdownify(h)
    '\\n1. item1\\n2. item2\\n\\n'

    Nested lists:
    
    >>> h = '<ol><li>item1</li><li>item2:<ul><li>item3</li><li>item4</li></ul></li></ol>'
    >>> tei2md.markdownify(h)
    '\\n1. item1\\n2. item2:\\n\\n\\t* item3\\n\\t* item4\\n\\t\\n\\n'

    Italics (<i> and <em> tags):

    >>> h = 'abc <em>def</em> ghi'
    >>> tei2md.markdownify(h)
    'abc *def* ghi'

    >>> h = 'abc <i>def</i> ghi'
    >>> tei2md.markdownify(h)
    'abc *def* ghi'


    Bold (<b> and <strong> tags):

    >>> h = 'abc <b>def</b> ghi'
    >>> tei2md.markdownify(h)
    'abc **def** ghi'

    >>> h = 'abc <strong>def</strong> ghi'
    >>> tei2md.markdownify(h)
    'abc **def** ghi'

    Tables:
    
    >>> h = '\
    <table>\
      <tr>\
        <th>th1aaa</th><th>th2</th>\
      </tr>\
      <tr>\
        <td>td1</td><td>td2</td>\
      </tr>\
    </table>'
    >>> tei2md.markdownify(h)
    '\\n\\n| th1aaa | th2 |\\n| ------ | --- |\\n| td1    | td2 |\\n\\n'
            
    i.e.:
        | th1aaa | th2 |
        | td1    | td2 |
"""

from bs4 import BeautifulSoup
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(path.dirname(root_folder)))
    sys.path.append(root_folder)

from openiti.new_books.convert.helper import html2md
from openiti.new_books.convert.helper.html2md import *  # import all constants!


class TeiConverter(html2md.MarkdownConverter):

    def convert(self, xml):
        """Convert xml to markdown.

        # We want to take advantage of the html5 parsing, but we don't actually
        # want a full document. Therefore, we'll mark our fragment with an id,
        # create the document, and extract the element with the id.
        """
        xml = html2md.wrapped % xml
        soup = BeautifulSoup(xml, "xml")
        #print(soup)
        text = self.process_tag(soup.find(id=FRAGMENT_ID), children_only=True)
        return self.post_process_md(text)

        

    #-----------------------------------------------------------

    # Additional conversion functions for specific tags (in alphabetic order):

    def convert_div1(self, el, text):
        """Converts tei xml <div1> tags to headings.

        Examples:
            >>> import tei2md
            >>> h = 'abc\
                     <div1 type="book" n="0" name="Preface">def</div1>\
                     ghi'
            >>> tei2md.markdownify(h)
            'abc\\n\\n### | [book 0: Preface]\\n\\ndef\\n\\nghi'
        """
        text = self.convert_div(el, text, n=1)
        return text

    def convert_div2(self, el, text):
        """Converts tei xml <div1> tags to headings.

        Examples:
            >>> import tei2md
            >>> h = 'abc\
                     <div2 type="book" n="0" name="Preface">def</div1>\
                     ghi'
            >>> tei2md.markdownify(h)
            'abc\\n\\n### || [book 0: Preface]\\n\\ndef\\n\\nghi'
        """
        text = self.convert_div(el, text, n=2)
        return text

    def convert_div3(self, el, text):
        """Converts tei xml <div1> tags to headings.

        Examples:
            >>> import tei2md
            >>> h = 'abc\
                     <div3 type="book" n="0" name="Preface">def</div1>\
                     ghi'
            >>> tei2md.markdownify(h)
            'abc\\n\\n### ||| [book 0: Preface]\\n\\ndef\\n\\nghi'
        """
        text = self.convert_div(el, text, n=3)
        return text

    def convert_div(self, el, text, n=None):
        """Converts tei xml <div> tags, depending on their type.
        
        Examples:
            >>> import tei2md
            >>> h = 'abc\
                     <div1 type="book" n="0" name="Preface">def</div1>\
                     ghi'
            >>> tei2md.markdownify(h)
            'abc\\n\\n### | [book 0: Preface]\\n\\ndef\\n\\nghi'

            >>> h = 'abc\
                     <div2 type="section" n="1">def</div1>\
                     ghi'
            >>> tei2md.markdownify(h)
            'abc\\n\\n### || [section 1]\\n\\ndef\\n\\nghi'

            >>> h = 'abc\
                     <div3 type="Aphorism">def</div1>\
                     ghi'
            >>> tei2md.markdownify(h)
            'abc\\n\\n### ||| [Aphorism]\\n\\ndef\\n\\nghi'

            Divs without type are stripped:
            
            >>> h = 'abc\
                     <div>def</div>\
                     ghi'
            >>> tei2md.markdownify(h)
            'abc def ghi'
        """
        if n == None:
            return text
            
        tag = "### " + "|"*n
        details ={"type": "", "n": "", "name": ""}
        for d in details:
            try:
                details[d] = el[d]
            except:
                pass
        title = "{} {}: {}".format(details["type"], details["n"], details["name"])
        title = re.sub("  |: $", " ", title).strip()
        return "\n\n{} [{}]\n\n{}\n\n".format(tag, title, text)

    def find_heading_level(self, el, text, level=None):
            return
        
    def convert_head(self, el, text, level=None):
        """Convert <head> tags to OpenITI mARkdown headings.

        NB: no general rule was found to define the heading level
        of the head tags. A default heading level of 3 has been assigned.
        For specific tei texts, a find_heading_level method
        could be defined to specify the heading level.        

        Examples:
            >>> import tei2md
            >>> h = 'abc\
                     <head>def</head>\
                     ghi'
            >>> tei2md.markdownify(h)
            'abc\\n\\n### ||| def\\n\\nghi'
        """
        if level == None:
            level = self.find_heading_level(el, text)
        if level == None:
            return "\n\n### ||| {}\n\n".format(text)
        else:
            return "\n\n### {} {}\n\n".format("|"*level, text)

    def convert_lb(self, el, text):
        """Convert tei <lb/> (line beginning) tags to newlines.

        Examples:
            >>> import tei2md
            >>> h = 'abc\
                     <lb/>def\
                     <lb/>ghi'
            >>> tei2md.markdownify(h)
            'abc\\n~~def\\n~~ghi'
        """
        return '\n~~'

    def convert_lg(self, el, text):
        """Convert TEI line groups (<lg>), which contain <l> elements.

        Examples:
            >>> import tei2md
            >>> h = '\
            abc\
            <lg>\
              <l>line1</l>\
              <l>line2</l>\
              <l>line3</l>\
              <l>line4</l>\
            </lg>\
            def'
            >>> tei2md.markdownify(h)
            ' abc\\n# line1\\n# line2\\n# line3\\n# line4\\n\\ndef'

        """
        lines = el.find_all("l", recursive=False)
        lines = [line.text for line in lines]
        return "\n# {}\n\n".format("\n# ".join(lines))
        
##    def convert_pb(self, el, text):
##        """Page beginnings are taken care of in a preprocessing step"""
##        pass

    def convert_quote(self, el, text):
        """Quotes are currently not converted"""
        return text


#######################################################################"

def preprocess_page_numbers(s):
    """Turn page beginnings into page endings.

    TEI xml indicates the beginning of a page, while OpenITI mARkdown
    indicates the end of a page.
    """
    try:
        s, tail = re.split("</body>", s)
    except:
        tail = ""
    s = re.split('<pb n="(\d+)"/?>', s)
    page_numbers = [it for it in s if s.index(it)%2]
    page_texts = [it for it in s if not s.index(it)%2]
    if page_texts[0] != "":
        page_numbers = ["0"] + page_numbers
    else:
        page_texts = page_texts[1:]
    text = ""
    for i, it in enumerate(page_texts):
        text += it
        text += "\n\nPageV00P{:03d}\n".format(int(page_numbers[i]))
    if tail:
        return text + "</body>" + tail
    else:
        return text
    
def preprocess_wrapped_lines(s):
    s = re.sub("-[\n\r]+~~", "", s)
    return re.sub("[\n\r]+~~", " ", s)


def markdownify(xml_str, **options):
    return TeiConverter(**options).convert(xml_str)


if __name__ == "__main__":
    test_xml = """
<?xml version="1.0" encoding="utf-8"?>
<TEI.2>
 <teiHeader>
 <fileDesc>
 <titleStmt>
 <title>Kitāb al-fihrist</title>
 <author>Ibn al-Nadīm</author>
 <editor>Gustav Flügel</editor>
 <funder>Andrew W. Mellon Foundation</funder>
 </titleStmt>
 <sourceDesc>
 <biblStruct>
 <monogr>
 <author>Ibn al-Nadīm</author>
 <title>Kitâb al-Fihrist</title>
 <editor>Gustav Flügel</editor>
 <imprint>
 <pubPlace>Leipzig</pubPlace>
 <publisher>Maktabat al-nahḍah al-miṣrīyah</publisher>
 <biblScope unit="vol">1</biblScope>
 <biblScope unit="pp" from="2" to="360">2-360</biblScope>
 <date>1871-1872</date>
 </imprint>
 </monogr>
 </biblStruct>
 </sourceDesc>
 </fileDesc>
 <encodingDesc>
 <editorialDecl default="NO">
 <correction status="high" default="NO" method="silent">
 <p>Data Entry</p>
 </correction>
 </editorialDecl>
 <refsDecl n="text=1" doctype="TEI.2">
 <state unit="text"/>
 <state unit="book"/>
 <state unit="chapter" n="chunk"/>
 </refsDecl>
 </encodingDesc>
 <profileDesc>
 <langUsage default="NO">
 <language id="arabic">Arabic</language>
 </langUsage>
 </profileDesc>
 </teiHeader>

<text lang="arabic" id="perseus0001.perseus001.alpheios-text-ara1">
<body>

<pb n="2"/>

<div1 type="book" n="1">
<div2 type="chapter" n="0" name="Preface">
<head>
الجزء الاول
من
كتاب الفهرست
</head>
<p>
بسم الله الرحمن الرحيم
</p>
<p>
رب يسر برحمتك النفوس تشراب الى النتائج دون المقدمات وترتاح الى الغرض المقصود
دون التطويل فى العبارات فلذلك اقتصرنا على هذه الكلمات فى صدر كتابنا هذه اذ كانت
دالة على ما قصدناه فى تاليفه ان شاء الله فنقول وبالله نستعين واياه نسئل الصلوة على
جميع انبيائه وعباده المخلصين فى طاعته ولا حول ولا قوة الا بالله العلى العظيم
</p>
<p>
هذا فهرست كتب جميع الامم من العرب والعجم الموجود منها بلغة العرب وقلمها
فى اصناف العلوم واخبار مصنفيها وطبقات مولفليها وانسابهم وتاريخ مواليدهم ومبلغ
اعمارهم واوقات وفاتهم واماكن بلدانهم ومناقبهم ومثالبهم منذ ابتاداء كل علم اخترع
الى عصرنا هذا وهو سنة سبع وسبعين وثلثمائة للهجرة
</p>
<head>
اقتصاص
ما يحتوى عليه الكتاب وهو عشر مقالات
</head>
<p>
المقالة الاولى وهى ثلثة فنون الفن الاول فى وصف لغات الامم من العرب
~~والعجم ونعوت
اقلامها وانواع خطوطها واشكال كتاباتها الفن الثانى فى اسماء كتب الشرائع
~~المنزلة على
مذاهب المسلمين ومذاهب اهلها الفن الثالث فى نعت الكتاب الذى لا ياتيه الباطل
من بين يديه ولا من خلفه تنزيل من حكيم حميد واسماء الكتب المصنفة فى علومه
واخبار القراء واسماء رواتهم والشواذ من قراءتهم
</p>

<pb n="3"/>

<p>
المقالة الثانية وهى ثلثة فنون فى النحويين واللغويين الفن الاول فى ابتداء
~~النحو واخبار
النحويين البصريين وفصحاء الاعراب واسماء كتبهم الفن الثانى فى اخبار النحويين
واللغويين من الكوفيين واسماء كتبهم الفن الثالث فى ذكر قوم من النحويين خلطوا
المذهبين واسماء كتبهم
</p>
</div2></div1></body></text></TEI.2>"""

    import doctest
    doctest.testmod()

    test_xml = preprocess_wrapped_lines(test_xml)
    test_xml = preprocess_page_numbers(test_xml)

    soup = BeautifulSoup(test_xml, "xml")
##    print(soup)
##    input()
##    print(soup.find_all("head"))
##    input()
    text_div = str(soup.find("text"))
##    print(text_div)
    text=markdownify(text_div, md_style="OpenITI")
##    print("*"*40)
    print(text)



"""Convert Noorlib library html to OpenITI mARkdown.

This script subclasses the generic MarkdownConverter class
from the html2md module (based on python-markdownify,
https://github.com/matthewwithanm/python-markdownify),
which uses BeautifulSoup to create a flexible converter.

The subclass in this module, NoorlibHtmlConverter,
adds methods specifically for the conversion of books from
the eShia library to OpenITI mARkdown:

* Span, div and p conversion: span, div and p classes needed to be converted
    are defined in self.class_dict.


Inheritance schema of the NoorlibHtmlConverter:

======================== ==========================
MarkdownConverter        NoorlibHtmlConverter
======================== ==========================
Options                  (inherited)
DefaultOptions           (inherited)
__init__                 (inherited)
__getattr__              (inherited)
convert                  (inherited)
process_tag              (inherited)
process_text             (inherited)
fill_out_columns         (inherited)
post_process_md          (inherited)
should_convert_tag       (inherited)
indent                   (inherited)
underline                (inherited)
create_underline_line    (inherited)
convert_a                (inherited)
convert_b                (inherited)
convert_blockquote       (inherited)
convert_br               (inherited)
convert_em               (inherited)
convert_hn               (inherited)
convert_i                (inherited)
convert_img              (inherited)
convert_list             (inherited)
convert_li               (inherited)
convert_ol               (inherited)
convert_p                convert_p
convert_table            (inherited)
convert_tr               (inherited)
convert_ul               (inherited)
convert_strong           (inherited)
                         convert_span
                         convert_div
======================== ==========================

"""
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(path.dirname(root_folder)))
    sys.path.append(root_folder)

from openiti.new_books.convert.helper import html2md
from openiti.new_books.convert.helper.html2md import *  # import all constants!


class NoorlibHtmlConverter(html2md.MarkdownConverter):
    """Convert Noorlib library html to OpenITI mARkdown.

    Examples:
        >>> import html2md_noorlib
        >>> h = '<img class="libimages" src="/images/books/86596/01/cover.jpg">'
        >>> html2md_noorlib.markdownify(h)
        '![](img/86596/01/cover.jpg)'

        >>> import html2md_noorlib
        >>> h = 'abc <a href="www.example.com">def</a> ghi'
        >>> html2md_noorlib.markdownify(h)
        'abc def ghi'
    """

    def __init__(self, **options):
        super().__init__(**options)
        self.class_dict = dict()
        self.class_dict["rightpome"] = "\n# {} %~% "              # <span class>
        self.class_dict["leftpome"] = "{}\n"                      # <span class>
        self.class_dict["footnote"] = "{}\n"                      # <div class>
##        ##old:
##        self.class_dict["Titr3"] = "\n\n### ||| {}\n\n"         # <span class>
##        self.class_dict["KalamateKhas2"] = "\n\n### || {}\n\n"  # <p class>
##        self.class_dict["KalamateKhas"] = "\n\n### ||| {}\n\n"  # <p class>
##        self.class_dict["TextsStyles3"] = "\n\n### ||| {}\n\n"  # <p class>
##        self.class_dict["TextsStyles1"] = "@QUR@ {}\n"          # <span class>
##        self.class_dict["Aye"] = "@QUR@ {}\n"                   # <span class>
##        self.class_dict["tdfehrest2"] = "\t{}"                  # <td class>
##        self.class_dict["list3"] = "\t{}"                       # <div class>
##        self.class_dict["sher"] = "# {}\n"                      # <p class>
##        self.class_dict["#6C3934"] = "\n\n# {}\n\n"             # <span class>

        self.options["image_link_regex"] = "/?images/books"
##        self.options["image_folder"] = "img"
        self.options["strip"] = ["a", "img"]


    def convert_span(self, el, text):
        """Converts html <span> tags, depending on their class attribute.

        Supported span classes should be stored in self.class_dict
        (key: span class (str); value: formatting string)
        E.g., {"quran": "@QUR@ {}\\n"}

        Example:
            >>> import html2md_noorlib
            >>> h = 'abc <span>def</span> ghi'
            >>> html2md_noorlib.markdownify(h)
            'abc def ghi'

            >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
            >>> html2md_noorlib.markdownify(h)
            'abc def ghi'

            #>>> h = 'abc <span class="Aya">def  ghi</span> jkl'
            #>>> html2md_noorlib.markdownify(h)
            #'abc @QUR02 def ghi jkl'

            # the @QUR@ example outputs are a result of post-processing;
            # the function itself will produce:
            # 'abc @QUR@ def ghi\\njkl'
            
            >>> h = '<span class="rightpome">abc def</span><span class="leftpome">ghi jkl</span>'
            >>> html2md_noorlib.markdownify(h)
            '\\n# abc def %~% ghi jkl'
        """
        try:  # will fail if el has no class attribute
            for c in el["class"]:
                #print(c)
                if c in self.class_dict:
                    return self.class_dict[c].format(text) if text else ''
                if c == "ayah":
                    try:
                        sura = el["surah"]
                    except:
                        sura = "0"
                    try:
                        aya = el["ayah"]
                    except:
                        aya = "0"
                    #print("@QUR{}.{}@ {}".format(sura, aya, text))
                    return "@QUR{}.{}@ {}\n".format(sura, aya, text)
        except Exception as e:
            pass
        return text


    def convert_div(self, el, text):
        """Converts html <div> tags, depending on their class attribute.

        Supported div classes should be stored in self.class_dict
        (key: div class (str); value: formatting string)

        Example:
            >>> import html2md_noorlib
            >>> h = 'abc <div>def</div> ghi'
            >>> html2md_noorlib.markdownify(h)
            'abc def ghi'

            >>> h = 'abc <div class="unknown_div_class">def</div> ghi'
            >>> html2md_noorlib.markdownify(h)
            'abc def ghi'

            >>> h = '<div class="ClssDivMeesage">Page Is Empty</div>'
            >>> html2md_noorlib.markdownify(h)
            ''
        """
        try:  # will fail if el has no class attribute
            for c in el["class"]:
                if c in self.class_dict:
                    return self.class_dict[c].format(text) if text else ''
                if c == "ClssDivMeesage":
                    return ""
        except Exception as e:
            pass
        return text

    def convert_p(self, el, text):
        """Converts <p> tags according to their class.

        Supported p classes should be stored in self.class_dict
        (key: span class (str); value: formatting string)
        E.g., {"quran": "@QUR@ {}\\n"}

        <p> tags without class attribute, or unsupported class,
        will be converted according to the markdown style
        as defined in the self.options["md_style"] value
        (from super().DefaultOptions)

        Examples:
            >>> import html2md_noorlib
            >>> h = "<p>abc</p>"
            >>> html2md_noorlib.markdownify(h)
            '\\n\\n# abc\\n\\n'

            >>> h = "<p>abc</p>"
            >>> html2md_noorlib.markdownify(h, md_style=ATX)
            '\\n\\nabc\\n\\n'

            >>> h = "<p></p>"
            >>> html2md_noorlib.markdownify(h, md_style=ATX)
            ''
        """
        if self.options['md_style'] == OPENITI:
            return '\n\n# %s\n\n' % text if text else ''
        else:
            return '\n\n%s\n\n' % text if text else ''

    def convert_sup(self, el, text):
        """Converts <sup> tags (used for footnote markers)."""
        return "({})".format(text.strip())
    


def markdownify(html, **options):
    """Shortcut to the convert method of the HindawiConverter class."""
    return NoorlibHtmlConverter(**options).convert(html)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""A converter for converting GRAR tei xml files into OpenITI format.

The converter has two main functions:
* convert_file: convert a single html file.
* convert_files_in_folder: convert all html files in a given folder

Usage examples:
    >>> from tei_converter_GRAR import convert_file, convert_files_in_folder
    >>> folder = r"test/GRAR/"
    >>> convert_file(folder+"GRAR000070.xml", dest_fp=folder+"converted/GRAR000070")
    >>> convert_files_in_folder(folder, dest_folder=folder+"converted")

Both functions use the GRARConverter class to do the heavy lifting.

The Graeco-Arabic studies website (graeco-arabic-studies.org)
contains 78 texts transcribed in TEI xml, and 21 additional
texts available only in html.

The XML texts were downloaded as is; for the html texts,
for each text every separate page was downloaded,
and a compound html file containing the metdata of that text
+ the div containing the text of each page was created
(using the combine_files_in_folder function).

The GRARConverter (which is sub-classed from tei_converter_generic.TeiConverter)
converts both the tei xml files and the html files.
It uses the generic TeiConverter's tei2md.TeiConverter for the xml files,
and for the html files the html2md_GRAR.GRARHtmlConverter
(which is a modified version (sub-class) of the html2md.MarkdownConverter).

Schema representing the method inheritance in the GRARConverter:

=========================== ========================== =========================
GenericConverter            TeiConverter               GRARConverter
=========================== ========================== =========================
__init__                    __init__ (appended)        (inherited)
convert_files_in_folder     (inherited)                convert_files_in_folder
convert file                convert_file               convert_file
make_dest_fp                (inherited)                (inherited)
get_metadata (dummy)        get_metadata               (inherited - for tei xml)
get_data                    (inherited)                (inherited - for tei xml)
pre_process                 pre_process                pre-process (appended)
add_page_numbers (dummy)    (inherited - not used)     (inherited - not used)
add_structural_annotations  add_structural_annotations (inherited - for tei xml)
remove_notes (dummy)        (inherited - generic!)     (inherited - generic!)
reflow                      (inherited)                (inherited)
add_milestones (dummy)      (inherited - dummy)        (inherited - not used)
post_process                post_process               post-process (appended)
compose                     (inherited)                (inherited)
save_file                   (inherited)                (inherited)
                            preprocess_page_numbers    (inherited)
                            preprocess_wrapped_lines   (inherited)
                                                       get_html_data
                                                       get_html_metadata
                                                       format_html_metadata
=========================== ========================== =========================

##Examples:
##    >>> from tei_converter_GRAR import GRARConverter
##    >>> conv = GRARConverter(dest_folder="test/GRAR/converted")
##    >>> conv.VERBOSE = False
##    >>> folder = r"test/GRAR"
##    >>> fn = r"GRAR000070.xml"
##    >>> conv.convert_file(os.path.join(folder, fn))

"""
import os
import re
from bs4 import BeautifulSoup

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.tei_converter_generic import TeiConverter
from openiti.new_books.convert.helper import html2md_GRAR



def convert_file(fp, dest_fp=None):
    """Convert one file to OpenITI format.

    Args:
        source_fp (str): path to the file that must be converted.
        dest_fp (str): path to the converted file. Defaults to None
            (in which case, the converted folder will be put in a folder
            named "converted" in the same folder as the source_fp)

    Returns:
        None
    """
    conv = GRARConverter() 
    conv.convert_file(fp, dest_fp=dest_fp)

def convert_files_in_folder(src_folder, dest_folder=None,
                            extensions=[], exclude_extensions=["yml"],
                            fn_regex=None):
    """Convert all files in a folder to OpenITI format.\
    Use the `extensions` and `exclude_extensions` lists to filter\
    the files to be converted.

    Args:
        src_folder (str): path to the folder that contains
            the files that must be converted.
        dest_folder (str): path to the folder where converted files
            will be stored.
        extensions (list): list of extensions; if this list is not empty,
            only files with an extension in the list should be converted.
        exclude_extensions (list): list of extensions;
            if this list is not empty,
            only files whose extension is not in the list will be converted.
        fn_regex (str): regular expression defining the filename pattern
            e.g., "-(ara|per)\d". If `fn_regex` is defined,
            only files whose filename matches the pattern will be converted.

    Returns:
        None
    """
    conv = GRARConverter()
    conv.convert_files_in_folder(src_folder, dest_folder=dest_folder,
                                 extensions=extensions,
                                 exclude_extensions=exclude_extensions,
                                 fn_regex=fn_regex)


################################################################################






class GRARConverter(TeiConverter):

    def pre_process(self, text):
        text = super().pre_process(text)
        text = re.sub("([\n ])[|*] ", r"\1", text)
        return text

    def format_html_metadata(self, meta_boxes):
        meta = {
            0: {
                "Domains": "", 
                "Title": "", 
            },
            1: {
                "Date": "", 
                "Translated from": "", 
                "Translator": "", 
                "Type": "", 
            },
            2: {
                "Author/Editor": "n.n.", 
                "Number": "", 
                "Pages": "", 
                "Publication type": "", 
                "Published": "n.d.", 
                "Publisher": "n.p.", 
                "Series": "", 
                "Title": "", 
                "Volume": "", 
            }
        }
        for i, box in enumerate(meta_boxes):
            for p in box.find_all("p"):
                #all_metadata[i].add(p.text.split(":")[0].strip())
                key = p.text.split(":")[0].strip()
                val = re.split(":", p.text, 1)[-1].strip()
                val = re.sub('English|Original| *" *|\n', "", val)
                val = re.sub(":", " :: ", val)
                meta[i][key] = val
     
        meta_strings = []
        meta_strings.append("#META# Title: {}".format(meta[0]["Title"]))
        meta_strings.append("#META# Genres: {}".format(meta[0]["Domains"]))
        for key in sorted(meta[1].keys(), reverse=True):
            if meta[1][key]:
                meta_strings.append("#META# {}: {}".format(key, meta[1][key]))
        ed_info = "#META# Ed_info: {} ({}), {}".format(meta[2]["Author/Editor"],
                                                       meta[2]["Published"],
                                                       meta[2]["Title"])
        if meta[2]["Series"]:
            ed_info += " ({} {}),".format(meta[2]["Series"], meta[2]["Volume"])
        ed_info += ", {}".format(meta[2]["Publisher"])
        if meta[2]["Volume"]:
            ed_info += ", vol. " + meta[2]["Volume"]
        if meta[2]["Pages"]:
            ed_info += ", pp. " + meta[2]["Pages"]
        meta_strings.append(ed_info)
        metadata = "\n".join(meta_strings)
        metadata = re.sub(" +", " ", metadata)
        metadata = re.sub(",+", ",", metadata)
        metadata = self.magic_value + metadata + self.header_splitter
        return metadata

    def get_html_metadata(self, soup):
        meta_boxes = soup.find_all("div", class_="boxcontent")
        return self.format_html_metadata(meta_boxes)

    def get_html_data(self, soup):
        text_soup = soup.find("body")
        c = html2md_GRAR.GRARHtmlConverter()
        text = c.convert(text_soup)
        return text

##    def convert_files_in_folder(self, folder):
##        for fn in os.listdir(folder):
##            fp = os.path.join(folder, fn)
##            if os.path.isfile(fp) and fn.endswith("ara1"):
##                self.convert_file(fp)

    def convert_file(self, source_fp, dest_fp=None):
        """Convert one file to OpenITI format.

        Args:
            source_fp (str): path to the file that must be converted.
            dest_fp (str): path to the converted file. Defaults to None
                (in which case, the converted folder will be put in a folder
                named "converted" in the same folder as the source_fp)

        Returns:
            None
        """
        if self.VERBOSE:
            print("converting", source_fp)
        if dest_fp == None:
            dest_fp = self.make_dest_fp(source_fp)
        with open(source_fp, mode="r", encoding="utf-8") as file:
            text = file.read()
        text = self.pre_process(text)
        soup = BeautifulSoup(text)
        metadiv = soup.find("div", class_="metadata")
        if metadiv: # if the file is a combined html file
            self.metadata = self.get_html_metadata(soup)
            metadiv.extract()
            text = self.get_html_data(soup)
        else: # if the file is a GRAR tei xml file:
            soup = BeautifulSoup(text, "xml")
            self.metadata = self.get_metadata(text)
            text = self.add_structural_annotations(text)

        text, notes = self.remove_notes(text)
        if not soup.find_all("lb") and not soup.find_all(title="linebreak"):
            text = self.reflow(text)
        text = self.post_process(text)
        text = self.compose(self.metadata, text, notes)

        self.save_file(text, dest_fp)

    def post_process(self, text, verbose=False):
        text = super().post_process(text)
        vols = re.findall("vol. (\d+) pp", self.metadata)
        #print(vols)
        if len(vols) == 1:
            text = re.sub("PageV00", "PageV{:02d}".format(int(vols[0])), text)
        elif len(vols) == 0:
            text = re.sub("PageV00", "PageV01", text)
        elif len(vols) > 1:
            print("More than one volume! adjust volume numbers manually")
        text = re.sub("# Page(V\d+P\d+)", r"Page\1\n\n#", text)
        if verbose:
            print(text)
            input()
        text = re.sub("\n~~ *(PageV\d+P\d+)\n", r"\n\n\1\n", text)
        text = re.sub("\n~~ ", r"\n~~", text)
        text = re.sub("(### \|+[^#]+)\n~~", r"\1 ", text)
        text = re.sub("(### \|+ \[[^\]]+\]) *\n+### \|{3}", r"\1", text)
        text = re.sub("([^.؟])\n{2}# (PageV\d+P\d+) ?", r"\1 \2\n~~", text)
        # if a page number does not follow a dot/question mark
        text = re.sub("([^.؟])\n{2}(PageV\d+P\d+)\n+# ", r"\1 \2\n~~", text)
        text = re.sub("([^.؟])\n{2}(PageV\d+P\d+) *(?!\n)", r"\1 \2\n~~", text)
        #text = re.sub("@QUOTE@", "", text)
        return text



###########################################################################


def combine_files_in_folder(folder, output_fp=None):
    """Combine separate html files into one master html file"""
    print(folder)
    comp = HTML_HEADER
    i = 0
    for fn in os.listdir(folder):
        if fn.endswith(".html") and not fn.endswith("combined.html"):
            fp = os.path.join(folder, fn)
            with open(fp, mode="r", encoding="utf-8") as file:
                html = file.read()
            soup = BeautifulSoup(html)
            if i == 0:
                meta_boxes = soup.find_all("div", class_="boxcontent")
                meta_boxes = [str(x) for x in meta_boxes]
                metadata = """<div class="metadata">\n{}\n</div>"""
                metadata = metadata.format("\n".join(meta_boxes))
                comp += metadata
            box = soup.find("div", class_="textbox")
            comp += "\n{}\n".format(box)
            i += 1
    comp += HTML_FOOTER
    comp = BeautifulSoup(comp).prettify()
    if not output_fp:
        #output_fp = os.path.join(folder, os.path.split(folder)[-1]+"_combined.html")
        #output_fp = re.findall("GRAR\d+", os.path.split(folder)[-1])[0]
        output_fp = os.path.join("combined", os.path.split(folder)[-1])
    with open(output_fp, mode="w", encoding="utf-8") as file:
        file.write(comp)

def list_all_tags(folder, header_end_tag="</teiHeader>"):
    """
    Extracts a list of all tags used in the texts in a folder:

    For GRAR:

    <body>
    <div1 type="" n="" (name="")(/)>     # book subdivision level 1
    <div2 type="" n="" (name="")(/)>     # book subdivision level 2 
    <head>                               # title
    <lb(/)>                              # start of new line
    <milestone unit="" n=""/>            #
    <p>                                  # paragraph
    <pb (type="") n=""/>                 # start of new page
    <quote type="" (author="") (id="")>  # quotation of a source
    <text lang="" id="">                 # metadata

    # tables: 
    <table>
    <tr>
    <td>

    # line groups (e.g., for poetry):
    <lg>                                 # line group
    <l>                                  # line in line group


    div types:

    ================= =================== 
    div1              div2
    ================= ===================
    book
    books
    chapter           chapter
    folio
    sentence          sentence
                      aphorism
    ================= ===================

    pb types: primary, secondary

    quote types: lemma, commentary

    milestone units: book, ed1chapter, ed1page, ms1folio
    """
    tags = []
    full_tags = []
    for fn in os.listdir(folder):
        fp = os.path.join(folder, fn)
        if not fn.endswith("yml") and not fn.endswith("py"):
            print(fn)
            with open(fp, mode="r", encoding="utf-8") as file:
                text = file.read()
    ##        orig_len = len(text)
            text = re.sub("\n~~", " ", text)
            if header_end_tag:
                text = re.sub(".+?{}".format(header_end_tag), "", text,
                              count=1, flags=re.DOTALL)
    ##        if "teiHeader" in text:
    ##            print(fn, "missing teiheader closing tag?")
    ##        if len(text) == orig_len:
    ##            print(fn, "missing teiheader?")
            text_full_tags = re.findall("<[^/][^>]*>", text)
            text_tags = re.findall("<[^/][^ >]+", text)
##            if '<milestone' in text_tags:
##                print("milestone in", fn)
##                input()
            for tag in set(text_tags):
                if tag not in tags:
                    tags.append(tag)
            for tag in set(text_full_tags):
                if tag not in full_tags:
                    full_tags.append(tag)

    stripped_tags = [re.sub('(author|lang|n|name|id)="[^"]+"', r'\1=""', tag)\
                     for tag in full_tags]
    stripped_tags = list(set(stripped_tags))

##    for tag in sorted(stripped_tags):
##        print(tag)
    for tag in sorted(full_tags):
        print(tag)


##############################################################################

if __name__ == "__main__":
##    conv = GRARConverter(dest_folder="test/converted")
##    conv.VERBOSE = False
##    folder = r"test"
##    fn = r"GRAR000070"
##    conv.convert_file(os.path.join(folder, fn))
##    input("passed test")
    import doctest
    doctest.testmod()

    input("Passed tests. Press Enter to start converting")
  

    folder = r"G:\London\OpenITI\RAWrabica\RAWrabica005000\GRAR"
    conv = GRARConverter(os.path.join(folder, "converted"))
    conv.extension = ""
    conv.convert_files_in_folder(folder)



"""This module provides useful regular expression patterns for OpenITI texts.

The patterns are roughly divided into the following sections:

# 1. Characters, words and spaces
# 2. OpenITI URIs and filenames
# 3. OpenITI mARkdown tags


See also:

* https://alraqmiyyat.github.io/mARkdown/
* https://docs.python.org/3/library/re.html
* https://pypi.org/project/regex/

"""

import re

# NB: "(?s)" = inline flag (to be put at the start of a regex pattern)
# that forces the regex machine to consider the dot as representing any
# character including newline (= using flag re.DOTALL while compiling a regex)
dotall = "(?s)"

# 1. Characters, words and spaces

ar_chars = """\
ء	ARABIC LETTER HAMZA
آ	ARABIC LETTER ALEF WITH MADDA ABOVE
أ	ARABIC LETTER ALEF WITH HAMZA ABOVE
ؤ	ARABIC LETTER WAW WITH HAMZA ABOVE
إ	ARABIC LETTER ALEF WITH HAMZA BELOW
ئ	ARABIC LETTER YEH WITH HAMZA ABOVE
ا	ARABIC LETTER ALEF
ب	ARABIC LETTER BEH
ة	ARABIC LETTER TEH MARBUTA
ت	ARABIC LETTER TEH
ث	ARABIC LETTER THEH
ج	ARABIC LETTER JEEM
ح	ARABIC LETTER HAH
خ	ARABIC LETTER KHAH
د	ARABIC LETTER DAL
ذ	ARABIC LETTER THAL
ر	ARABIC LETTER REH
ز	ARABIC LETTER ZAIN
س	ARABIC LETTER SEEN
ش	ARABIC LETTER SHEEN
ص	ARABIC LETTER SAD
ض	ARABIC LETTER DAD
ط	ARABIC LETTER TAH
ظ	ARABIC LETTER ZAH
ع	ARABIC LETTER AIN
غ	ARABIC LETTER GHAIN
ـ	ARABIC TATWEEL
ف	ARABIC LETTER FEH
ق	ARABIC LETTER QAF
ك	ARABIC LETTER KAF
ل	ARABIC LETTER LAM
م	ARABIC LETTER MEEM
ن	ARABIC LETTER NOON
ه	ARABIC LETTER HEH
و	ARABIC LETTER WAW
ى	ARABIC LETTER ALEF MAKSURA
ي	ARABIC LETTER YEH
ً	ARABIC FATHATAN
ٌ	ARABIC DAMMATAN
ٍ	ARABIC KASRATAN
َ	ARABIC FATHA
ُ	ARABIC DAMMA
ِ	ARABIC KASRA
ّ	ARABIC SHADDA
ْ	ARABIC SUKUN
٠	ARABIC-INDIC DIGIT ZERO
١	ARABIC-INDIC DIGIT ONE
٢	ARABIC-INDIC DIGIT TWO
٣	ARABIC-INDIC DIGIT THREE
٤	ARABIC-INDIC DIGIT FOUR
٥	ARABIC-INDIC DIGIT FIVE
٦	ARABIC-INDIC DIGIT SIX
٧	ARABIC-INDIC DIGIT SEVEN
٨	ARABIC-INDIC DIGIT EIGHT
٩	ARABIC-INDIC DIGIT NINE
ٮ	ARABIC LETTER DOTLESS BEH
ٰ	ARABIC LETTER SUPERSCRIPT ALEF
ٹ	ARABIC LETTER TTEH
پ	ARABIC LETTER PEH
چ	ARABIC LETTER TCHEH
ژ	ARABIC LETTER JEH
ک	ARABIC LETTER KEHEH
گ	ARABIC LETTER GAF
ی	ARABIC LETTER FARSI YEH
ے	ARABIC LETTER YEH BARREE
۱	EXTENDED ARABIC-INDIC DIGIT ONE
۲	EXTENDED ARABIC-INDIC DIGIT TWO
۳	EXTENDED ARABIC-INDIC DIGIT THREE
۴	EXTENDED ARABIC-INDIC DIGIT FOUR
۵	EXTENDED ARABIC-INDIC DIGIT FIVE"""
##‌	ZERO WIDTH NON-JOINER
##‍	ZERO WIDTH JOINER"""
ar_chars = "".join([x.split("\t")[0] for x in ar_chars.splitlines()])
ar_char = "[{}]".format(ar_chars) # regex for one Arabic character
ar_tok = "[{}]+".format(ar_chars) # regex for one Arabic token

noise = re.compile(""" ّ    | # Tashdīd / Shadda
                       َ    | # Fatḥa
                       ً    | # Tanwīn Fatḥ / Fatḥatān
                       ُ    | # Ḍamma
                       ٌ    | # Tanwīn Ḍamm / Ḍammatān
                       ِ    | # Kasra
                       ٍ    | # Tanwīn Kasr / Kasratān
                       ْ    | # Sukūn
                       ۡ    | # Quranic Sukūn
                       ࣰ    | # Quranic Open Fatḥatān
                       ࣱ    | # Quranic Open Ḍammatān
                       ࣲ    | # Quranic Open Kasratān
                       ٰ    | # Dagger Alif
                       ـ     # Taṭwīl / Kashīda
                   """, re.VERBOSE)

any_unicode_letter = "[^\W\d_]"
any_word = any_unicode_letter + "+"

###### regex module (alternative to re module):
#any_unicode_letter = "\p{L}"
#any_unicode_diacritic = "\p{M}"

space ="(?: |[\r\n]+~~|PageV[^P]+P\d+)+"
space ="(?:\W|PageV[^P]+P\d+)+" # also takes care of commas etc.
space_word = "(?:{}{})".format(space, ar_tok)


# 2. URIs and OpenITI filenames:

# OpenITI URIs
language_codes = [        # ISO 639-2B language codes: https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes
    "ara", # Arabic
    "per", # Persian
    "urd", # Urdu
    "heb", # Hebrew
    "arm", # Armenian
    "ave", # Avestan
    "geo", # Georgian
    "kur", # Kurdish
    "pus", # Pashtu
    "swa", # Swahili
    ]
auth = r"\b\d{4}[A-Z][a-zA-Z]+"
book = auth + "\.[A-Z][a-zA-Z]+"
version = book + "\.\w+(?:Vols)?(?:BK\d+|[A-Z])*-(?:%s)+\d+" % "|".join(language_codes)
author_uri = auth
book_uri = book
version_uri = version

# OpenITI text file names:
extensions = ["inProgress", "completed", "mARkdown"]
ext_regex = r"(?:\.{}|(?= |\n|\r|\Z))".format("|\.".join(extensions))
version_file = version+ext_regex
version_fp = r"%s[/\\]%s[/\\]%s" % (auth, book, version_file)

# OpenITI yml file names:
yml ="\.yml"
auth_yml = auth+yml
book_yml = book+yml
version_yml = version+yml

auth_yml_fp = r"%s[/\\]%s" % (auth, auth_yml)
book_yml_fp = r"%s[/\\]%s[/\\]%s" % (auth, book, book_yml)
version_yml_fp = r"%s[/\\]%s[/\\]%s" % (auth, book, version_yml)

# 3. OpenITI mARkdown tags

magic_value = "######OpenITI#"
header_splitter = "#META#Header#End#"

# Page numbers:
vol_page = "PageV[^P]+P\d+[AB]?"
vol_no = "PageV([^P]+)P\d+[AB]?"
page_no = "PageV[^P]+P(\d+[AB]?)"
vol_page_3 = "PageV[^P]{2}P\d{3}[AB]?"
vol_page_4 = "PageV[^P]{2}P\d{4}[AB]?"
page = dotall + r"(?:(?<={})|(?<={})|(?<={})).+?(?:{}|\Z)".format(vol_page_3,
                    vol_page_4, header_splitter, vol_page)

# Hierarchical section tags:
section_tag = "### \|+ "
section_title = section_tag + "([^\r\n]*)"
section = dotall + section_tag + r".+?(?=###|\Z)"
section_text = dotall + section_tag + "[^\r\n]*[\r\n]+(.+?)(?=###|\Z)"

# biographies
bio_tag = r"### (?:\|+ )?\$+ "
bio = dotall + bio_tag + "[^\r\n]*[\r\n]+.+?(?=###|\Z)"
bio_title = bio_tag + "([^\r\n]*)"
bio_text = dotall + bio_tag + "[^\r\n]*[\r\n]+(.+?)(?=###|\Z)"

bio_man_tag = "### (?:\|+ )?\$ "
bio_man = dotall + bio_man_tag + "[^\r\n]*[\r\n]+.+?(?=###|\Z)"
bio_man_title = bio_man_tag + "([^\r\n]*)"
bio_man_text = dotall + bio_man_tag + "[^\r\n]*[\r\n]+(.+?)(?=###|\Z)"

bio_woman_tag = "### (?:\|+ )?\$\$ "
bio_woman = dotall + bio_woman_tag + "[^\r\n]*[\r\n]+.+?(?=###|\Z)"
bio_woman_title = bio_woman_tag + "([^\r\n]*)"
bio_woman_text = dotall + bio_woman_tag + "[^\r\n]*[\r\n]+(.+?)(?=###|\Z)"

# editorial sections:
editorial_tag = "### \|EDITOR\|"
editorial = dotall + editorial_tag + r"[^\r\n]*[\r\n]+.+?(?=###|\Z)"
editorial_text = dotall + editorial_tag + "[^\r\n]*[\r\n]+(.+?)(?=###|\Z)"

# paratext sections:
paratext_tag = "### \|PARATEXT\|"
paratext = dotall + paratext_tag + r"[^\r\n]*[\r\n]+.+?(?=###|\Z)"
paratext_text = dotall + paratext_tag + r"[^\r\n]*[\r\n]+(.+?)(?=###|\Z)"

# paragraphs:
paragraph_tag = r"(?<=[\r\n])# "
paragraph = r"(?<=[\r\n])# [^#]+"
paragraph_text = r"(?<=[\r\n]# )[^#]+"
word_in_paragraph = r"(?<=[\r\n])# [^#]+?{}[^#]+" # insert word using string formatting
word_in_paragraph_text= r"(?<=[\r\n]# )[^#]+?{}[^#]+" # insert word using string formatting

# years:
year = r"\bY[A-Z]?\d+"
year_born = r"\bYB\d+"
year_died = r"\bYD\d+"

# analytical tag pattern:
anal_tag = "(?:@[A-Z]{3})?[@#][A-Z]{3}(?:\$[\w+\-]+)?(?:@?\d\d+)?"
tag_range = "|".join([str(i)+"%(w)s{"+str(i)+"}" for i in range(1,21)])
tag_range = "@?(?:\d\W*?(?:" + tag_range % {"w": space_word} + "))?"
anal_tag_text = "(?:@[A-Z]{3})?[@#][A-Z]{3}(?:\$[\w+\-]+)?" + tag_range

#(this is identical to:)
#anal_tag_text = """(?x)                                 # verbose inline flag
#                   (?:@[A-Z]{3})?                       # optional personal ID
#                   [@#][A-Z]{3}                         # tag name/category
#                   (?:_[a-zA-Z]+)*                      # subcategory/ies
#                   (?:$[\w\-]+)?                        # named entity ID
#                   @?(?:\d                              # tag range: prefix
#                     (?:1%(w)s{1}|2%(w)s{2}|3%(w)s{3}|  # tag range: no. of words
#                     4%(w)s{4}|5%(w)s{5}|6%(w)s{6}|
#                     7%(w)s{7}|8%(w)s{8}|9%(w)s{9}|
#                     10%(w)s{10}|11%(w)s{11}|12%(w)s{12}|
#                     13%(w)s{13}|14%(w)s{14}|15%(w)s{15}|
#                     16%(w)s{16}|17%(w)s{17}))?""" % {"w": space_word}

if __name__ == "__main__":
    # tests for the regex patterns involved:
    verbose=True
    def test_regex_findall(ptrn, txt, res):
        shouldbe = "Should be {}".format(res)
        if verbose:
            print(re.findall(ptrn, txt))
        assert re.findall(ptrn, txt) == res, shouldbe

    # test word/character regexes:
    
    test_regex_findall(ar_char, "كتاب kitab", ['ك', 'ت', 'ا', 'ب'])
    test_regex_findall(ar_tok, "كتاب kitab", ['كتاب'])
    res = ['ك', 'ت', 'ا', 'ب', 'k', 'i', 't', 'a', 'b']
    test_regex_findall(any_unicode_letter, "كتاب_kitab", res)
    test_regex_findall(any_word, "كتاب_kitab", ['كتاب', 'kitab'])
    txt = """الكلمة الأولى
~~السطر الثاني PageV01P003 الصفحة التالية"""
    res = [' الأولى', '\n~~السطر', ' الثاني', ' PageV01P003 الصفحة', ' التالية']
    test_regex_findall(space_word, txt, res)

    # test URI/fn regexes:
    
    fn = "0255Jahiz.Hayawan.Shamela0001234VolsBk1-ara1.completed"
    test_regex_findall(auth, fn, ["0255Jahiz"])
    test_regex_findall(book, fn, ["0255Jahiz.Hayawan"])
    test_regex_findall(version, fn,
                       ["0255Jahiz.Hayawan.Shamela0001234VolsBk1-ara1"])
    test_regex_findall(version_file, fn,
                       ["0255Jahiz.Hayawan.Shamela0001234VolsBk1-ara1.completed"])
    fn = "0255Jahiz.Hayawan.Shamela0001234-per3"
    test_regex_findall(version, fn,
                       ["0255Jahiz.Hayawan.Shamela0001234-per3"])
    test_regex_findall(version_file, fn,
                       ["0255Jahiz.Hayawan.Shamela0001234-per3"])
    fps = """
0255Jahiz/0255Jahiz.yml
0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.yml
0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Shamela0001234-per3.inProgress
0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Shamela0001234-per3.yml
"""
    test_regex_findall(auth_yml, fps, ["0255Jahiz.yml"])
    test_regex_findall(book_yml, fps, ["0255Jahiz.Hayawan.yml"])
    test_regex_findall(version_yml, fps, ["0255Jahiz.Hayawan.Shamela0001234-per3.yml"])
    test_regex_findall(auth_yml_fp, fps,
                       ["0255Jahiz/0255Jahiz.yml"])
    test_regex_findall(book_yml_fp, fps,
                       ["0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.yml"])
    test_regex_findall(version_yml_fp, fps,
                       ["0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Shamela0001234-per3.yml"])
    test_regex_findall(version_fp, fps,
                       ["0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Shamela0001234-per3.inProgress"])
    
    # test OpenITI mARkdown regexes:

    txt = """######OpenITI#
#META# blabla
#META#Header#End#

Text text text"""
    test_regex_findall(header_splitter, txt, ["#META#Header#End#"])
    test_regex_findall(magic_value, txt, ["######OpenITI#"])
    pages = "PageV01P001 PageV02P0002 PageVM3P003 Pagev04P004 PageV05p0005"
    test_regex_findall(vol_page, pages, ["PageV01P001", "PageV02P0002", "PageVM3P003"])
    test_regex_findall(vol_no, pages, ["01", "02", "M3"])
    test_regex_findall(page_no, pages, ["001", "0002", "003"])
    pages = "#META#Header#End#\n\npage text 1 PageV01P001 page text 2 PageV02P0002 text without page number"
    res = ['\n\npage text 1 PageV01P001', ' page text 2 PageV02P0002', ' text without page number']
    test_regex_findall(page, pages, res)
    
    txt = """### | heading 1
text section 1
### || heading 2
text section 2
text section 2"""
    test_regex_findall(section_tag, txt, ["### | ", "### || "])
    test_regex_findall(section_title, txt, ["heading 1", "heading 2"])
    test_regex_findall(section, txt, ['### | heading 1\ntext section 1\n',
                                      '### || heading 2\ntext section 2\ntext section 2'])
    test_regex_findall(section_text, txt, ['text section 1\n',
                                           'text section 2\ntext section 2'])
    
    txt = """### | biographies
text section 1
### || $ biography 1
text biography 1
### $$ biography 2
text biography 2"""
    test_regex_findall(bio_tag, txt, ["### || $ ", "### $$ "])
    test_regex_findall(bio_title, txt, ["biography 1", "biography 2"])
    test_regex_findall(bio, txt, ["### || $ biography 1\ntext biography 1\n",
                                  "### $$ biography 2\ntext biography 2"])
    test_regex_findall(bio_text, txt, ["text biography 1\n",
                                       "text biography 2"])

    
    test_regex_findall(bio_man_tag, txt, ["### || $ "])
    test_regex_findall(bio_man_title, txt, ["biography 1"])
    test_regex_findall(bio_man, txt, ["### || $ biography 1\ntext biography 1\n"])
    test_regex_findall(bio_man_text, txt, ["text biography 1\n"])
    

    test_regex_findall(bio_woman_tag, txt, ["### $$ "])
    test_regex_findall(bio_woman_title, txt, ["biography 2"])
    test_regex_findall(bio_woman, txt, ["### $$ biography 2\ntext biography 2"])
    test_regex_findall(bio_woman_text, txt, ["text biography 2"])


    txt = """
### |EDITOR|
editorial intro
### | title
main text
### |EDITOR|
editorial outro"""
    test_regex_findall(editorial_tag, txt, ["### |EDITOR|", "### |EDITOR|"])
    test_regex_findall(editorial, txt, ["### |EDITOR|\neditorial intro\n",
                                        "### |EDITOR|\neditorial outro"])
    test_regex_findall(editorial_text, txt, ["editorial intro\n",
                                             "editorial outro"])

    txt = """
### |PARATEXT|
paratext intro
### | title
main text
### |PARATEXT|
paratext outro"""
    test_regex_findall(paratext_tag, txt, ["### |PARATEXT|", "### |PARATEXT|"])
    test_regex_findall(paratext, txt, ["### |PARATEXT|\nparatext intro\n",
                                        "### |PARATEXT|\nparatext outro"])
    test_regex_findall(paratext_text, txt, ["paratext intro\n",
                                             "paratext outro"])
    
    txt = """### | section title
# paragraph 1
~~paragraph 1 continued
# paragraph 2
### | section 2
# paragraph 3"""
    test_regex_findall(paragraph_tag, txt, ["# ", "# ", "# "])
    test_regex_findall(paragraph, txt, ["# paragraph 1\n~~paragraph 1 continued\n",
                                        "# paragraph 2\n",
                                        "# paragraph 3"])
    test_regex_findall(paragraph_text, txt, ["paragraph 1\n~~paragraph 1 continued\n",
                                             "paragraph 2\n",
                                             "paragraph 3"])
    w = word_in_paragraph.format("continued")
    test_regex_findall(w, txt, ["# paragraph 1\n~~paragraph 1 continued\n"])
    w = word_in_paragraph_text.format("continued")
    test_regex_findall(w, txt, ["paragraph 1\n~~paragraph 1 continued\n"])

    txt = """wulida YB0100 sana mia wa-mata YD170 sana mia wa-sabcin"""
    test_regex_findall(year, txt, ["YB0100", "YD170"])
    test_regex_findall(year_born, txt, ["YB0100"])
    test_regex_findall(year_died, txt, ["YD170"])

    txt = "مع #PER$0310Tabari أبي جرير الطبري #PER$0279Baladhuri@12 وأحمد البلاذري"
    test_regex_findall(anal_tag, txt, ["#PER$0310Tabari", "#PER$0279Baladhuri@12"])
    test_regex_findall(anal_tag_text, txt, ["#PER$0310Tabari", "#PER$0279Baladhuri@12 وأحمد البلاذري"])


    txt = """مع @MGR@PER$0310Tabari أبي جرير الطبري @PER$0279Baladhuri@12 وأحمد\n~~البلاذري"""
    test_regex_findall(anal_tag, txt, ["@MGR@PER$0310Tabari", "@PER$0279Baladhuri@12"])
    test_regex_findall(anal_tag_text, txt, ["@MGR@PER$0310Tabari",
                                            "@PER$0279Baladhuri@12 وأحمد\n~~البلاذري"])

    txt = """قال: @QUR$1_1-2@08 ( بسم الله الرحمن الرحيم الحمد لله رب العلمين )"""
    test_regex_findall(anal_tag, txt, ["@QUR$1_1-2@08"])
    test_regex_findall(anal_tag_text, txt, ["@QUR$1_1-2@08 ( بسم الله الرحمن الرحيم الحمد لله رب العلمين"])


    print("finished testing")




import math
import os
import random
import re
import unicodedata
import urllib.request as url
import requests

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.dirname(path.abspath(__file__))))
    sys.path.append(root_folder)
from openiti.helper import ara

splitter = "#META#Header#End#"
milestone = "Milestone300"
thresh = 1000

exclude_folders = ["OpenITI.github.io", "Annotation", "maintenance",
                   "i.mech00", "i.mech01", "i.mech02", "i.mech03",
                   "i.mech04", "i.mech05", "i.mech06", "i.mech07",
                   "i.mech", "i.mech_Temp", "i.mech08", "i.mech09",
                   "i.logic", "i.cex", "i.cex_Temp", ".git"]

exclude_files = ["README.md", ".DS_Store",
                 ".gitignore", "text_questionnaire.md"]

def get_all_text_files_in_folder(start_folder, excluded_folders=exclude_folders,
                                 exclude_files=exclude_files):
    """A generator that yields the file path for all OpenITI text files \
    in a folder and its subfolders.

    OpenITI text files are defined here as files that have a language
    identifier (-ara1, -ara2, -per1, etc.) and have either no extension
    or .mARkdown, .completed, or .inProgress.
    
    The script creates a generator over which you can iterate.
    It yields the full path to each of the text files.

    Args:
        start_folder (str): path to the folder containing the text files
        excluded_folders (list): list of folder names that should be excluded
            (default: the list of excluded folders defined in this module)
        excluded_files (list): list of file names that should be excluded
            (default: the list of excluded file names defined in this module)

    Examples:
        > folder = r"D:\London\OpenITI\25Y_repos"
        > for fp in get_all_text_files_in_folder(folder):
            print(fp)
        > folder = r"D:\London\OpenITI\25Y_repos\0025AH"
        > AH0025_file_list = [fp for fp in get_all_text_files_in_folder(folder)]
    """
    for root, dirs, files in os.walk(start_folder):
        dirs[:] = [d for d in dirs if d not in exclude_folders]
        files[:] = [f for f in files if f not in exclude_files]
        for fn in files:
            if re.findall(r"-\w\w\w\d(?:.inProgress|.completed|.mARkdown)?\Z", fn):
                fp = os.path.join(root, fn)
                yield(fp)

#def get_all_yml_files_in_folder(start_folder, yml_type,
def get_all_yml_files_in_folder(start_folder, yml_types,
                                excluded_folders=exclude_folders,
                                exclude_files=exclude_files):
    """A generator that yields the file path for all yml files \
    of a specific type in a folder and its subfolders.

    OpenITI yml files exist for authors, books and versions.
    
    The script creates a generator over which you can iterate.
    It yields the full path to each of the yml files.

    Args:
        start_folder (str): path to the folder containing the text files
        yml_type (list): list of desired yml file types:
            one or more of "author", "book", or "version"
        excluded_folders (list): list of folder names that should be excluded
            (default: the list of excluded folders defined in this module)
        excluded_files (list): list of file names that should be excluded
            (default: the list of excluded file names defined in this module)

    Examples:
        > folder = r"D:\London\OpenITI\25Y_repos"
        > for fp in get_all_yml_files_in_folder(folder):
            print(fp)
        > folder = r"D:\London\OpenITI\25Y_repos\0025AH"
        > AH0025_file_list = [fp for fp in get_all_text_files_in_folder(folder)]
    """
    dots = {"author": 1, "book": 2, "version": 3}
    if isinstance(yml_types, str):
        yml_types = [yml_types,]
    for root, dirs, files in os.walk(start_folder):
        dirs[:] = [d for d in dirs if d not in exclude_folders]
        files[:] = [f for f in files if f not in exclude_files]
        for fn in files:
            for yml_type in yml_types:
                if re.findall(r"^(?:[^.]+\.){%s}yml$" % dots[yml_type], fn):
                    fp = os.path.join(root, fn)
                    yield(fp)


def get_all_characters_in_text(fp):
    """Get a set of all characters used in a text.

    Args:
        fp (str): path to a text file.

    Returns:
        (set): a set of all characters used in the text.
    """
    with open(fp, mode="r", encoding="utf-8") as file:
        text = file.read()
        return set(text)
    

def get_all_characters_in_folder(start_folder, verbose=False,
                                 exclude_folders=[], exclude_files=[]):
    """Get a set of all characters used in all OpenITI text files \
    in a folder and its subfolders.

    Args:
        start_folder (str): path to the root directory. All files and folders
            in it, except if they are in the exclude lists, will be processed.
        verbose (bool): if True, filenames and current number of characters
            in the set will be printed.
        exclude_folders (list): list of folder names to be excluded
            from the process.
        exclude_folders (list): list of file names to be excluded.

    Returns:
        (set): a set of all characters used in the folder.
    """
    all_characters = set()
    for root, dirs, files in os.walk(start_folder):
        dirs[:] = [d for d in dirs if d not in exclude_folders]
        files[:] = [f for f in files if f not in exclude_files]
        for fn in files:
            fp = os.path.join(root, fn)
            extensions = [".completed", ".mARkdown", ".inProgress"]
            if os.path.splitext(fn)[1] in extensions \
              or re.findall(r"(ara|per)\d$", fn):
                if verbose:
                    print(len(all_characters), fn)
                text_chars = get_all_characters_in_text(fp)
                all_characters = all_characters.union(text_chars)
    return all_characters


def get_character_names(characters, verbose=False):
    """Print the unicode name of a list/set/string of characters.

    Args:
        characters (list/set/string): a list, string or set of characters.
        verbose (bool): if set to True, the output will be printed

    Returns:
        (dict): a dictionary of characters and their names.

    Examples:
        >>> char_dict = {"١": "ARABIC-INDIC DIGIT ONE",\
                         "٢": "ARABIC-INDIC DIGIT TWO"}
        >>> char_dict == get_character_names("١٢")
        True
        >>> char_dict == get_character_names(["١", "٢"])
        True
        >>> char_dict == get_character_names({"١", "٢"})
        True
    """
    char_dict = dict() 
    for c in sorted(list(characters)):
        try:
            name = unicodedata.name(c)            
        except:
            name = None
        char_dict[c] = name
        if verbose:
            print("{}\t{}".format(c, name))
    
    return char_dict

def text_cleaner(text):
    """Clean text by normalizing Arabic characters \
    and removing all non-Arabic characters

    Args:
        text (str): the string to be cleaned

    Returns:
        (str): the cleaned string
    """
    text = ara.normalize_ara_light(text)
    text = re.sub("\W|\d|[A-z]", " ", text)
    text = re.sub(" +", " ", text)
    return text


def roundup(x, par):
    new_x = int(math.ceil(int(x) / float(par)) * par)
    return new_x


def generate_ids_through_permutations(char_string_for_ids, id_len_char, limit):
    list_char = list(char_string_for_ids) * id_len_char

    dic = {}
    iterations = 0

    while len(dic) < limit:
        new = "".join(random.sample(list_char, id_len_char))
        dic[new] = 0

        iterations += 1
        if iterations % 100000 == 0:
            print("\tITERATIONS: %d; DICTIONARY: %d" % (iterations, len(dic)))

    ids = "\n".join(list(dic.keys()))

    # where `L` is the length of IDs, `T` is the total number of unique IDs.
    file_name = "IDs_ASCII_L%d_T%d.txt" % (id_len_char, len(dic))
    with open(file_name, 'w', encoding='utf8') as outfile:
        outfile.write(ids)

    print("=" * 80)
    print("Generating TXT file with unique IDs, based on:")
    print("\tPermuations (L=%d) of: %s" % (id_len_char, char_string_for_ids))
    print("\tTotal number of IDs: %s" % '{:,}'.format(len(dic)))
    print("=" * 80)



def read_header(pth, lines=300, header_splitter="#META#Header#End#", 
                encoding="utf-8-sig"):
    """Read only the OpenITI header of a file without opening the entire file.

    Args:
        pth (str): path to local text file / URL of remote text file 
        lines (int): number of lines at the top of the file to be read
        header_splitter (str): string that separates the header from the body text
        encoding (str): text encoding to use. Default: "utf-8-sig" 
            (Unicode utf-8, strips BOM at start of file)

    Returns:
        (str): the metadata header of the text file
    """
    header = ""
    i=0
    try:   # local text file:
        with open(pth, mode="r", encoding=encoding) as file:
            while i <= lines:
                # read next line in file:
                line = file.readline()
                # add the line to the header:
                header += line
                # stop and return the header when we reach the header splitter:
                if header_splitter in line:
                    return header
                # start a new iteration
                i += 1
    except:  # URL of online text:
        with requests.get(pth, stream=True) as r:
            for line in r.iter_lines():
                # decode the new line and add it to the header:
                line = line.decode(encoding)
                header += line+"\n"   # r.iter_lines() strips off the newline character!
                # stop and return the header when we reach the header splitter:
                if header_splitter in line:
                    return header
                # start a new iteration iteration unless we run out of lines to read:
                i+=1
                if i >= lines:
                    break

    print("{}: header splitter not reached after {} lines".format(fp, lines))
    
    return ""




def read_text(pth, max_header_lines=300, split_header=False, remove_header=False,
              encoding="utf-8-sig", header_splitter="#META#Header#End#"):
    """Read a text from a file or from a URL.
    
    The parameters allow you to choose to  
    * full text file content: metadata header + text in a single string
    * only the text, without the header, in a single string (remove_header=True)
    * header and text, separated, in a tuple of strings (split_header=True)

    Args:
        pth (str): path to local text file / URL of remote text file 
        max_header_lines (int): number of lines at the top of the file to be read to find the header
        split_header (bool): if True, the header and main body of the text 
            will be returned as separate strings
        remove_header (bool): if True, only the main body of the text will be returned
        encoding (str): text encoding to use. Defaults to "utf-8-sig" 
            (Unicode utf-8, strips BOM at start of file)
        header_splitter (str): string that separates the header from the body text.
            Defaults to "#META#Header#End#" (end of the standard OpenITI metadata header)

    Returns: 
        str or tuple
    """
    # full text+header:
    if not split_header and not remove_header:
        try:
            with open(pth, mode="r", encoding=encoding) as file:
                return file.read()
        except:
            r = requests.get(pth)
            return r.text
    # split the main text from the header:
    else:
        # get the header:
        header = read_header(pth, lines=max_header_lines)
        
        # get the main body of the text:
        try:   # path to local text file:
            with open(pth, mode="r", encoding="utf-8") as file:
                # read the full text file:
                text = file.read()
                # strip the header:
                text = text[len(header):]
                
        except:  # URL to online text:
            with requests.get(pth) as r:
                # download and read the full text file:
                text = r.text
                # strip the header:
                text = text[len(header):]
        
        if remove_header:  # return only the main body of the text
            return text
        # else: return both header and text
        return (header, text) 


def absolute_path(path):
    return os.path.abspath(path)

def get_page_number(page_numbers, pos):
    """Get the page number of a token at index position `pos` in a string \
    based on a dictionary `page_numbers` that contains the index positions \
    of the page numbers in that string.

    Args:
        page_numbers (dict):
            key: index of the last character of the page number in the string
            value: page number
        pos (int): the index position of the start of a token in the string
    """
    for k in sorted(page_numbers.keys()):
        if pos < k:
            return page_numbers[k]

def report_missing_numbers(fp, no_regex="### \$ \((\d+)",
                           report_repeated_numbers=True):
    """Use a regular expression to check whether numbers\
    (of books, pages, etc.) are in sequence and no numbers are missing.

    Arguments:
        fp (str): path to the text file
        no_regex (str): regular expression pattern describing the number
            for which the sequence should be checked.
            NB: the numbers should be in the first/only capture group

    Use cases:
        - Page numbers: use regex `PageV\d+P(\d+)`
        - numbered sections: e.g.,
          `### \$ \(?(\d+)` for dictionary items,
          `### \|{2} (\d+)` for second-level sections, ...
    """
    with open(fp, mode="r", encoding="utf-8") as file:
        text = file.read()
    current_num = 0
    page_numbers = {m.end(): m.group(0) \
                    for m in re.finditer("PageV\d+P\d+", text)}
    for match in re.finditer(no_regex, text):
        no = int(match.group(1))
        if no == 1:
            current_num = 1
            page = get_page_number(page_numbers, match.start())
            print("start recounting from 1 at", page)
        elif no == current_num:
            page = get_page_number(page_numbers, match.start())
            if report_repeated_numbers:
                print(page, no, "follows", current_num)
        elif no != current_num + 1:
            page = get_page_number(page_numbers, match.start())
            print(page, no, "follows", current_num)
            current_num = no
        else:
            current_num = no

def natural_sort(obj):
    """Sort a list containing letters and numbers in its natural order
    (1,2,3,4,5,6,7,8,9,10, ... instead of 1,10,2,3,4,5,6,7,8,9,10)

    based on https://stackoverflow.com/a/16090640/4045481
    """
    natsort = lambda s: [int(t) if t.isdigit() else t.lower()
                         for t in re.split('(\d+)', s)]
    return sorted(obj, key=natsort)
    

def get_semantic_tag_elements(tag_name, text, include_tag=False,
                                  include_prefix=False, include_offsets=False,
                                  max_tokens=99, normalize_spaces=False):
    """Extract semantic tags (the likes of @TOP\d\d+) from OpenITI texts

    Args:
        tag_name (str): the tag you want to extract (e.g., @TOP, @PER, ...)
        text (str): the string from which the tags are to be extracted
        include_tag (bool): if False, only the content of the tag
            will be returned. If True, both tag+content (default: False)
        include_prefix (bool): if False, the prefix (that is, the number
            of characters defined by the first digit after the tag)
            will be stripped off from the result. Only if include_tag is
            set to False. Default: False.
        include_offsets (bool): if True, the start and end offsets of
            each element will be included (as a dictionary:
            with keys "match", "start", "end")
        max_tokens (int): the maximum number of tokens inside a tag.
            Default: 99.
        normalize_spaces (bool): if True, new lines, page numbers etc. 
            will be removed from the returned tokens.

    Returns:
        list
    """
    # first extract a large number of tokens after the tag
    ar_tok = "[" + "".join(ara.ar_chars)+"]+"
    not_token_not_tag = "[^@"+ar_tok[1:]
    n_tokens = "{1,"+str(max_tokens)+"}"
    #pattern = tag_name+"\d\d+(?:[^@\w]+\w+){1,"+str(max_tokens)+"}"
    pattern = f"{tag_name}\d\d+(?:{not_token_not_tag}{ar_tok}){n_tokens}"
    tmp_results = re.finditer(pattern, text)

    # select the amount of tokens as defined in the tag:
    final_results = []
    for result in tmp_results:
        # extract the tag from the result:
        res_tag = re.findall(tag_name+"\d\d+", result.group())[0]
        not_token = "[^"+ar_tok[1:]
        tokens = re.split("("+not_token+")", result.group()[len(res_tag):])
        # remove empty string matches:
        tokens = [tok for tok in tokens if tok != ""]
        n_prefix, n_toks = re.findall("(\d)(\d+)", res_tag)[-1]
        if not include_tag:
            selected_toks = tokens[1:(2*int(n_toks))]
            cleaned_res = "".join(selected_toks)
            # move start offset to first character of the content of the tag:
            start_offset = result.start() + len(res_tag+tokens[0])
            if not include_prefix:
                # remove the prefix from the match:
                cleaned_res = cleaned_res[int(n_prefix):]
                # move the start offset to the first character after the prefix:
                start_offset += int(n_prefix)
            end_offset =  start_offset + len(cleaned_res)
            # remove new lines, page numbers, etc. from the tokens:
            if normalize_spaces:
                selected_toks = [re.sub(not_token, " ", tok) for tok in selected_toks]
                cleaned_res = "".join(selected_toks)
                if not include_prefix:
                    # remove the prefix from the match:
                    cleaned_res = cleaned_res[int(n_prefix):]
            if include_offsets:
                final_results.append({"match":cleaned_res,
                                      "start":start_offset,
                                      "end":end_offset})
            else:
                final_results.append(cleaned_res)
        else:
            selected_toks = tokens[0:(2*int(n_toks))]
            cleaned_res = res_tag+("".join(selected_toks))
            start = result.start()
            end = start+len(cleaned_res)
            if normalize_spaces:
                selected_toks = [re.sub(not_token, " ", tok) for tok in selected_toks]
                cleaned_res = res_tag+("".join(selected_toks))
            if include_offsets:
                final_results.append({"match":cleaned_res,
                                      "start":start,
                                      "end":end})
            else:
                final_results.append(cleaned_res)

    return final_results


def find_section_title(loc, section_titles, section_starts):
    """Find the section title(s) for a character offset

    Args:
        loc (int): character offset for which the section title is wanted
        section_titles (list): a list of all section titles in the document
        section_starts (list): a list of character offsets of the
            starts of all sections in the text
    """
    i = bisect(section_starts, loc)
    return section_titles[i]

def get_sections(text, section_header_regex="### .+", include_hierarchy=True):
    """Get the section titles and start offsets for all sections in the text

    Args:
        text (str): the text containing the sections
        section_header_regex (str): regular expression pattern for section headers
        include_hierarchy (bool): if False, only the title of the section
            will be returned; if True, a list of titles of all parent sections
            will be returned

    Returns:
        list (section_titles, section_starts)
    """
    section_titles = [] 
    section_starts = [] 
    open_sections = [] 
    for m in re.finditer(section_header_regex, text):
        title = m.group()
        level = title.count("|") - 1
        if not include_hierarchy:
            open_sections = title
        elif level == 0:
            open_sections = [title,] 
        else:
            open_sections = open_sections[:level]
            open_sections.append(title)
         
        section_titles.append(open_sections) 
        section_starts.append(m.start())

    return section_titles, section_starts


if __name__ == "__main__":
    import doctest
    doctest.testmod()

    start_folder = r"D:\London\OpenITI\25Y_repos"
    chars = get_all_characters_in_folder(start_folder, verbose=False,
                                         exclude_folders=exclude_folders,
                                         exclude_files=exclude_files)
    #get_character_names(chars, verbose=True)
    test_str = """هذه محاولة ثانية: @TOP01 بغداد مدينة
يعرف أيضا @TOP12 بمدينة ms001 PageV01P001 السلام.
واسم المؤسس: @PER01 أحمد!"""


    res = get_semantic_tag_elements("@T(?:OP)?", test_str, include_tag=False, include_prefix=False, include_offsets=True)
    print(res)



"""A converter for converting LAL tei xml files into OpenITI format.

(LAL = Library of Arabic Literature)

The converter has two main functions:
* convert_file: convert a single html file.
* convert_files_in_folder: convert all html files in a given folder

Usage examples:
    >>> from tei_converter_LAL import convert_file, convert_files_in_folder
    >>> folder = r"test/LAL/"
    >>> convert_file(folder+"LAL000070.xml", dest_fp=folder+"converted/LAL000070")
    >>> convert_files_in_folder(folder, dest_folder=folder+"converted")

Both functions use the LALConverter class to do the heavy lifting.



The LALConverter (which is sub-classed from tei_converter_generic.TeiConverter)
converts both the tei xml files and the html files.
It uses the generic TeiConverter's tei2md.TeiConverter for the xml files,
and for the html files the html2md_LAL.LALHtmlConverter
(which is a modified version (sub-class) of the html2md.MarkdownConverter).

Schema representing the method inheritance in the LALConverter:

=========================== ========================== =========================
GenericConverter            TeiConverter               LALConverter
=========================== ========================== =========================
__init__                    __init__ (appended)        (inherited)
convert_files_in_folder     (inherited)                convert_files_in_folder
convert file                convert_file               convert_file
make_dest_fp                (inherited)                (inherited)
get_metadata (dummy)        get_metadata               (inherited - for tei xml)
get_data                    (inherited)                (inherited - for tei xml)
pre_process                 pre_process                pre-process (appended)
add_page_numbers (dummy)    (inherited - not used)     (inherited - not used)
add_structural_annotations  add_structural_annotations (inherited - for tei xml)
remove_notes (dummy)        (inherited - generic!)     (inherited - generic!)
reflow                      (inherited)                (inherited)
add_milestones (dummy)      (inherited - dummy)        (inherited - not used)
post_process                post_process               post-process (appended)
compose                     (inherited)                (inherited)
save_file                   (inherited)                (inherited)
                            preprocess_page_numbers    (inherited)
                            preprocess_wrapped_lines   (inherited)
                                                       get_html_data
                                                       get_html_metadata
                                                       format_html_metadata
=========================== ========================== =========================

##Examples:
##    >>> from tei_converter_LAL import LALConverter
##    >>> conv = LALConverter(dest_folder="test/LAL/converted")
##    >>> conv.VERBOSE = False
##    >>> folder = r"test/LAL"
##    >>> fn = r"LAL000070.xml"
##    >>> conv.convert_file(os.path.join(folder, fn))

"""
import os
import re
from bs4 import BeautifulSoup

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.tei_converter_generic import TeiConverter
from openiti.new_books.convert.helper import html2md_LAL
from openiti.helper.ara import deNoise



def convert_file(fp, dest_fp=None):
    """Convert one file to OpenITI format.

    Args:
        source_fp (str): path to the file that must be converted.
        dest_fp (str): path to the converted file. Defaults to None
            (in which case, the converted folder will be put in a folder
            named "converted" in the same folder as the source_fp)

    Returns:
        None
    """
    conv = LALConverter() 
    conv.convert_file(fp, dest_fp=dest_fp)

def convert_files_in_folder(src_folder, dest_folder=None,
                            extensions=[], exclude_extensions=["yml"],
                            fn_regex=None):
    """Convert all files in a folder to OpenITI format.\
    Use the `extensions` and `exclude_extensions` lists to filter\
    the files to be converted.

    Args:
        src_folder (str): path to the folder that contains
            the files that must be converted.
        dest_folder (str): path to the folder where converted files
            will be stored.
        extensions (list): list of extensions; if this list is not empty,
            only files with an extension in the list should be converted.
        exclude_extensions (list): list of extensions;
            if this list is not empty,
            only files whose extension is not in the list will be converted.
        fn_regex (str): regular expression defining the filename pattern
            e.g., "-(ara|per)\d". If `fn_regex` is defined,
            only files whose filename matches the pattern will be converted.

    Returns:
        None
    """
    conv = LALConverter()
    conv.convert_files_in_folder(src_folder, dest_folder=dest_folder,
                                 extensions=extensions,
                                 exclude_extensions=exclude_extensions,
                                 fn_regex=fn_regex)


################################################################################






class LALConverter(TeiConverter):


    def extract_meta_item(self, tei_header, tag_name, labels_d, meta):
        """Extract author, title, editor from teiHeader"""
        items = tei_header.titleStmt.find_all(tag_name)
        for item in items:
            #lang = item["xml:lang"]
            lang = item["xml:id"].split(".")[0]
            try:
                label = labels_d[tag_name][lang]
            except:
                label = tag_name[0].upper() + tag_name[1:]
            try:
                meta += "#META# {}: {}\n".format(label, item.text.strip())
            except Exception as e:
                print(tag_name, "not found:", e)
                pass
        return meta

    def extract_publ_stmt(self, tei_header, meta):
        """Extract the publicationStmt info from the teiHeader"""
        try:
            isbn = tei_header.publicationStmt.idno
            meta += "#META# ISBN: {}\n".format(isbn.text.strip())
        except Exception as e:
            print("isbn not found:", e)
            pass
        try:
            publisher = tei_header.publicationStmt.publisher
            meta += "#META# Publisher: {}\n".format(publisher.text.strip())
        except Exception as e:
            print("Publisher not found:", e)
            pass
        try:
            pubPlace = tei_header.publicationStmt.pubPlace
            meta += "#META# Place of publication: {}\n".format(pubPlace.text.strip())
        except Exception as e:
            print("pubPlace not found:", e)
            pass
        return meta

    def extract_witnesses(self, tei_header, meta):
        """Extract witness data from the teiHeader"""
        # try:
        #     witnesses = tei_header.sourceDesc.listBibl
        # except Exception as e:
        #     try: 
        #         witnesses = tei_header.sourceDesc
        #     except:
        #         print("witnesses not found:", e)
        #         return meta
        # for w in witnesses.find_all("bibl"):
        #     w = w.text.strip()
        #     meta += "#META# Witness: {}\n".format(w)
        try:
            witnesses = tei_header.sourceDesc.listBibl
            for w in witnesses.find_all("bibl"):
                w = w.text.strip()
                meta += "#META# Witness: {}\n".format(w)
            return meta
        except:
            pass

        for ref in tei_header.find_all("bibl"):
            if "xml:id" in ref.attrs and "wit" in ref["xml:id"]:
                w = ref.text.strip()
                meta += "#META# Witness: {}\n".format(w)
        return meta
                
    def extract_ed_decl(self, tei_header, meta):
        """Extract the editorial declaration from the teiHeader"""
        try:
            decl = tei_header.editorialDecl
            for p in decl.find_all("p"):
                p = re.sub("[\r\n]+", " ", p.text.strip())
                meta += "#META# Editorial declaration: {}\n".format(p)
        except Exception as e:
            print("editorial declaration not found:", e)
            pass
        return meta    

    def get_metadata(self, text):
        """Extracts the metadata from the text.


        Args:
            text (str): text of the file that contains data + metadata

        Returns:
            metadata (str): metadata formatted in OpenITI format
                (including the magic value and the header splitter)
        """
        soup = BeautifulSoup(text, "xml")
        labels_d = {
            "title":  {"ARA": "العنوان", "ENG": "Title (EN)"},
            "author": {"ARA": "المؤلف", "ENG": "Author"},
            "editor": {"ARA": "المحقق", "ENG": "Editor"},
            }
        meta = ""
        tei_header = soup.teiHeader
        meta = self.extract_meta_item(tei_header, "title", labels_d, meta)
        meta = self.extract_meta_item(tei_header, "author", labels_d, meta)
        meta = self.extract_meta_item(tei_header, "editor", labels_d, meta)
        meta = self.extract_publ_stmt(tei_header, meta)
        meta = self.extract_witnesses(tei_header, meta)
        meta = self.extract_ed_decl(tei_header, meta)
        meta = re.sub("([a-z])([A-Z])", r"\1 \2", meta)
        meta = re.sub("\n+(?!#META#)", " ", meta)
        meta = "######OpenITI#\n\n" + meta + "\n\n#META#Header#End#\n\n"
        
        return meta

    def pre_process(self, text):
        text = super().pre_process(text)
        # remove zero-width joiner:
        text = re.sub(r"‌", "", text)
        text = re.sub(" *<caesura/> *", " %~% ", text)
        text = re.sub("[\r\n]+", " ", text)
##        soup = BeautifulSoup(text, "xml")
##        # remove the front page:
##        soup.extract("front")
##        # remove notes:
##        soup.extract("note")
##        # convert caesura tags into OpenITI poetry tags:
##        for caesura in soup.find_all("caesura"):
##            caesura.replace_with(" %~% ")
##        text = soup.prettify()
        return text


##    def convert_files_in_folder(self, folder):
##        for fn in os.listdir(folder):
##            fp = os.path.join(folder, fn)
##            if os.path.isfile(fp) and fn.endswith("ara1"):
##                self.convert_file(fp)

    def add_structural_annotations(self, text, **options):
        soup = BeautifulSoup(text, "xml")
        text_soup = soup.find("body")
        text = html2md_LAL.markdownify(str(text_soup), **options)
        return text

    def convert_file(self, source_fp, dest_fp=None):
        """Convert one file to OpenITI format.

        Args:
            source_fp (str): path to the file that must be converted.
            dest_fp (str): path to the converted file. Defaults to None
                (in which case, the converted folder will be put in a folder
                named "converted" in the same folder as the source_fp)

        Returns:
            None
        """
        if self.VERBOSE:
            print("converting", source_fp)
        if dest_fp == None:
            dest_fp = self.make_dest_fp(source_fp)
        with open(source_fp, mode="r", encoding="utf-8") as file:
            text = file.read()
        text = self.pre_process(text)
        self.metadata = self.get_metadata(text)
        text = self.add_structural_annotations(text, strip=["front", "note"])

        # notes already removed by stripping <note> elements:
        notes = ""
        #text, notes = self.remove_notes(text)
        
        text = deNoise(text)

        soup = BeautifulSoup(text, "xml")
        if not soup.find_all("lb") and not soup.find_all(title="linebreak"):
            text = self.reflow(text)
        
        text = self.post_process(text)
        
        text = self.compose(self.metadata, text, notes)

        self.save_file(text, dest_fp)

    def post_process(self, text, verbose=False):
        text = super().post_process(text)
        vols = re.findall("vol. (\d+) pp", self.metadata)
        #print(vols)
        if len(vols) == 1:
            text = re.sub("PageV00", "PageV{:02d}".format(int(vols[0])), text)
        elif len(vols) == 0:
            text = re.sub("PageV00", "PageV01", text)
        elif len(vols) > 1:
            print("More than one volume! adjust volume numbers manually")
        text = re.sub("# Page(V\d+P\d+)", r"Page\1\n\n#", text)

        text = re.sub("\n~~ *(PageV\d+P\d+)\n", r"\n\n\1\n", text)
        text = re.sub("\n~~ ", r"\n~~", text)
        text = re.sub("(### \|+[^#]+)\n~~", r"\1 ", text)
        text = re.sub("(### \|+ \[[^\]]+\]) *\n+### \|{3}", r"\1", text)
        text = re.sub("([^.؟])\n{2}# (PageV\d+P\d+) ?", r"\1 \2\n~~", text)
        # if a page number does not follow a dot/question mark
        text = re.sub("([^.؟])\n{2}(PageV\d+P\d+)\n+# ", r"\1 \2\n~~", text)
        text = re.sub("([^.؟])\n{2}(PageV\d+P\d+) *(?!\n)", r"\1 \2\n~~", text)
        #text = re.sub("@QUOTE@", "", text)
        text = re.sub("(### \|PARATEXT\|(?:\n# .+)+)\n### \|PARATEXT\|", r"\1", text)
        # fix sections without headers:
        text = re.sub("(### \|+)([\r\n]+)", r"\1 \2", text)
        # put poems always on one line:
        text = re.sub(r"(%~%.+)[\r\n]+~~", r"\1 ", text)
        # remove waṣlas:
        text = re.sub("ٱ", "ا", text)
        # Format the paragraph numbers as page number:
        # (this proved to be the wrong approach: some text have a number for every line)
        # temp = ""
        # current_p = ""
        # for p in re.split("(\[[\d.]+\])", text):
        #     if p.startswith("["):
        #         split_p = p[1:-1].split(".")
        #         if len(split_p) == 2:
        #             v, p = split_p
        #         elif len(split_p) > 2:
        #             v = split_p[0]
        #             p = "0".join(split_p[1:])
        #             print(split_p)
        #         else:
        #             v = 0
        #             p = split_p[0]
        #         current_p = "PageV{:02d}P{:03d}".format(int(v), int(p))
        #     else:
        #         temp += p + current_p
        # text = re.sub("\[(\d)\]", r"PageV00P00\1", temp)
        # text = re.sub("\[(\d\d)\]", r"PageV00P0\1", text)
        # text = re.sub("\[(\d\d+)\]", r"PageV00P\1", text)

        return text



###########################################################################


def list_all_tags(folder, header_end_tag="</teiHeader>"):
    """
    Extracts a list of all tags used in the texts in a folder:


    """
    tags = []
    full_tags = []
    for fn in os.listdir(folder):
        fp = os.path.join(folder, fn)
        #if not fn.endswith("yml") and not fn.endswith("py"):
        if fn.endswith("xml"):
            with open(fp, mode="r", encoding="utf-8") as file:
                text = file.read()
    ##        orig_len = len(text)
            text = re.sub("\n~~", " ", text)
            if header_end_tag:
                text = re.sub(".+?{}".format(header_end_tag), "", text,
                              count=1, flags=re.DOTALL)
    ##        if "teiHeader" in text:
    ##            print(fn, "missing teiheader closing tag?")
    ##        if len(text) == orig_len:
    ##            print(fn, "missing teiheader?")
            text_full_tags = re.findall("<[^/][^>]*>", text)
            text_tags = re.findall("<[^/][^ >]+", text)
##            if '<milestone' in text_tags:
##                print("milestone in", fn)
##                input()
            for tag in set(text_tags):
                if tag not in tags:
                    tags.append(tag)
            for tag in set(text_full_tags):
                if tag not in full_tags:
                    full_tags.append(tag)

    stripped_tags = [re.sub('(author|lang|n|name|id)="[^"]+"', r'\1=""', tag)\
                     for tag in full_tags]
    stripped_tags = list(set(stripped_tags))

##    for tag in sorted(stripped_tags):
##        print(tag)
    for tag in sorted(full_tags):
        print(tag)


##############################################################################

if __name__ == "__main__":
##    conv = LALConverter(dest_folder="test/converted")
##    conv.VERBOSE = False
##    folder = r"test"
##    fn = r"LAL000070"
##    conv.convert_file(os.path.join(folder, fn))
##    input("passed test")
    import doctest
    doctest.testmod()

    #input("Passed tests. Press Enter to start converting")
  

    folder = r"C:\Users\peter\Downloads\LAL"
    conv = LALConverter(os.path.join(folder, "converted"))
    #fn = "al-Quḍāʿī (ed. Qutbuddin) A Treasury of Virtues.xml"
    #conv.convert_file(os.path.join(folder, fn))
    #input("CONTINUE?")
    conv.extension = ""
    conv.VERBOSE = True
    conv.convert_files_in_folder(folder)



"""Converter that converts HTML files from the Rafed library
to OpenITI mARkdown.

The converter has two main functions:
* convert_file: convert a single html file.
* convert_files_in_folder: convert all html files in a given folder

Usage examples:
    >>> from html_converter_Rafed import convert_file
    >>> folder = r"test/Rafed/"
    >>> convert_file(folder+"10584.html", dest_fp=folder+"converted/10584")
    >>> from html_converter_Rafed import convert_files_in_folder
    >>> convert_files_in_folder(folder, dest_folder=folder+"converted")  

Both functions use the RafedHtmlConverter class to do the heavy lifting.
The RafedHtmlConverter is a subclass of GenericHtmlConverter,
which in its turn inherits many functions from the GenericConverter.

GenericConverter
    \_ GenericHtmlConverter
            \_ EShiaHtmlConverter
            \_ RafedHtmlConverter
            \_ ...

Overview of the methods of these classes:
(methods of GenericConverter are inherited by GenericHtmlConverter;
and methods of GenericHtmlConverter are inherited by RafedHtmlConverter.
Methods of the superclass with the same name
in the subclass are overwritten by the latter)

================================== ========================== ==========================
GenericConverter                   GenericHtmlConverter       RafedHtmlConverter
================================== ========================== ==========================
__init__                           __init__                   (inherited)
convert_files_in_folder            (inherited)                (inherited)
convert file                       (inherited)                (inherited)
make_dest_fp                       (inherited - generic!)     (inherited - generic!)
get_metadata (dummy)               (inherited - dummy!)       get_metadata
get_data                           get_data                   (inherited)
pre_process                        (inherited)                pre_process
add_page_numbers (dummy)           (inherited - dummy!)       add_page_numbers
add_structural_annotations (dummy) add_structural_annotations add_structural_annotations
remove_notes (dummy)               remove_notes               remove_notes
reflow                             (inherited)                (inherited)
add_milestones (dummy)             (inherited - dummy!)       (inherited - dummy!)
post_process                       (inherited - generic!)     post_process
compose                            (inherited)                (inherited)
save_file                          (inherited)                (inherited)
                                   inspect_tags_in_html       (inherited)
                                   inspect_tags_in_folder     (inherited)
                                   find_example_of_tag        (inherited)
================================== ========================== ==========================

The RafedHtmlConverter's add_structural_annotations method uses html2md_Rafed,
an adaptation of the generic html2md (based on markdownify)
to convert the html tags to OpenITI annotations. 

Examples:
    >>> from html_converter_Rafed import RafedHtmlConverter
    >>> conv = RafedHtmlConverter()
    >>> conv.dest_folder = r"test/Rafed/converted"
    >>> conv.VERBOSE = False
    >>> folder = r"test/Rafed/"
    >>> conv.convert_file(folder+"Rafed_3943.html")
    >>> conv.convert_files_in_folder(folder, ["html"])
"""

from bs4 import BeautifulSoup
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.html_converter_generic import GenericHtmlConverter
from openiti.new_books.convert.helper import html2md_Rafed


def convert_file(fp, dest_fp=None):
    """Convert one file to OpenITI format.

    Args:
        source_fp (str): path to the file that must be converted.
        dest_fp (str): path to the converted file. Defaults to None
            (in which case, the converted folder will be put in a folder
            named "converted" in the same folder as the source_fp)

    Returns:
        None
    """
    conv = RafedHtmlConverter() 
    conv.convert_file(fp, dest_fp=dest_fp)

def convert_files_in_folder(src_folder, dest_folder=None,
                            extensions=["html"], exclude_extensions=["yml"],
                            fn_regex=None):
    """Convert all files in a folder to OpenITI format.\
    Use the `extensions` and `exclude_extensions` lists to filter\
    the files to be converted.

    Args:
        src_folder (str): path to the folder that contains
            the files that must be converted.
        dest_folder (str): path to the folder where converted files
            will be stored.
        extensions (list): list of extensions; if this list is not empty,
            only files with an extension in the list should be converted.
        exclude_extensions (list): list of extensions;
            if this list is not empty,
            only files whose extension is not in the list will be converted.
        fn_regex (str): regular expression defining the filename pattern
            e.g., "-(ara|per)\d". If `fn_regex` is defined,
            only files whose filename matches the pattern will be converted.

    Returns:
        None
    """
    conv = RafedHtmlConverter()
    conv.convert_files_in_folder(src_folder, dest_folder=dest_folder,
                                 extensions=extensions,
                                 exclude_extensions=exclude_extensions,
                                 fn_regex=fn_regex)


################################################################################


class RafedHtmlConverter(GenericHtmlConverter):

    def get_metadata(self, text):
        soup = BeautifulSoup(text)
        meta_div = soup.find_all("div", class_="bookinfo")[0]
        meta = html2md_Rafed.markdownify(meta_div)
        meta = [line.strip() for line in meta.splitlines() if len(line) > 3]
        meta = "#META# " + ("\n#META# ".join(meta))
        meta = self.magic_value + meta + self.header_splitter
        return meta
        

    def pre_process(self, text):
        """Remove superfluous elements from the html file before processing."""

        def remove_html_elements(soup, tag_name, class_=None, contains_str=None):
            """Remove all html elements with tag `tag` and class `class_` \
            if they contain `contains_str`
            
            Args:
                soup (BeautifulSoup object): BeautifulSoup representation
                    of the html file
                tag_name (str): name of the tag that needs to be removed
                    (e.g., "p", "div", "span"). 
                class_ (str): class of the tag that needs to be removed.
                    Defaults to None. If None, all `tag_name` elements
                    will be removed, regardless of their class.
                contains_str (str): defaults to None. If not None,
                    `tag_name` tags will only be removed if the text within them
                    contains the `contains_str` string.
            """
            soup = BeautifulSoup(text)
            soup = soup.find("div", {"id": "content1"})
            if class_:
                elements = soup.find_all(tag_name, class_=class_)
                #print(len(elements), tag_name, "elements with class", class_, "will be removed")
            else:
                elements = soup.find_all(tag_name)
            for el in elements:
                if contains_str:
                    if contains_str in el.text:
                        el.extract()
                else:
                    el.extract()
            return soup.prettify()
        
        text = super().pre_process(text)
        
        # attach separated wa- and a- prefixes to the following word: 
        text = re.sub(r"\b([وأ])[\s~]+", r"\1", text)

        # remove superfluous html elements: 
        #soup = BeautifulSoup(text)
        text = remove_html_elements(text, "p", class_="rfdLine")
        text = remove_html_elements(text, "script")
##        remove_html_elements(soup, "TITLE")
##
##        for fn_line in soup.find_all("hr", class_="content_hr"):
##            fn_line.insert_after("FOOTNOTES")
##        text = soup.prettify()
        
        return text


    def add_page_numbers(self, text, source_fp, vol=None):
        """Convert the page numbers in the text into OpenITI mARkdown format

        In Rafed texts, the page numbers are at the bottom of the page,
        in a div with class="pgnum">.
        Volume numbers are not mentioned in the html files,
        but every volume is a different html file.

        It also deletes the page header after extracting the page number.
        """

        # try to get the volume number from the filename:
        try:
            vol_no = "PageV{:02d}P{}".format(vol, "{:03d}")
        except:
            vol_no = "PageV01P{:03d}"

        # add the page number
        soup = BeautifulSoup(text)
        for div in soup.find_all("div", class_="pgnum"):
            div_text = div.text.strip()
            page_no = re.findall("\d+", div_text)[-1]
            # add the page number string to the text:
            div.insert_after(vol_no.format(int(page_no)))
            # remove the original page number div:
            div.extract()
        return soup.prettify()


    def remove_notes(self, text):
        """Remove footnotes from text, and format them as endnotes.

        Footnotes in Rafed html files are below a horizontal line
        (<HR class=content_hr>), located just below the page number
        (<P class=content_paragraph><SPAN class=content_text>ص:45</SPAN></P></DIV></DIV>)
        each footnote in a <DIV id=content_note_PAGE_NOTENUMBER class=content_note>(footnote text)</DIV>
        
        This function extracts the footnotes from the texts
        and turns them into endnotes.
        The markers that indicate the location of the notes
        within the text are not removed.
        """
        split_text = re.split("(PageV\d+P\d+)", text)
        text = []
        footnotes = []
        for i, t in enumerate(split_text):
            if re.match("PageV\d+P\d+", t):
                text.append(t)
            else:
                page = re.findall('<div class="pgcontent">.+', t, flags=re.DOTALL)
                if not page:
                    #print("NO PAGE CONTENT FOUND!")
                    #print(t)
                    continue
                soup = BeautifulSoup(page[0])
                #print(page)
                #print(soup.prettify())
                note_ps = soup.find_all("p", class_="rfdFootnote0")
                #print(len(note_ps))
                notes = []
                if note_ps:
                    #print("removing notes")
                    for note in note_ps:
                        notes.append(note.text.strip())
                        # remove the note from the html string:
                        note.extract()
                    try:
                        notes = "\n".join(notes) + "\n" +  split_text[i-1] + "\n\n"
                    except:
                        notes = "\n".join(notes) + "\n" + "PageV00P000\n\n"
                    footnotes.append(notes)
                
                #input("CONTINUE?")
                text.append(soup.prettify())

        text = "\n\n".join(text)
        notes = "\n\n".join(footnotes)
        notes = re.sub("\n+#* *\n+", "\n\n", notes)
        notes = self.endnote_splitter + notes
        return text, notes 
                    

    def add_structural_annotations(self, html):
        """Convert html to mARkdown text using a html2md converter."""
        text = html2md_Rafed.markdownify(html)
        return text             


    def post_process(self, text):
        """Deal with formatting problems introduced during the conversion process."""
        text = super().post_process(text)

        # remove page numbers of empty pages:
        text = re.sub("(PageV\d+P\d+)\s*PageV\d+P\d+", r"\1", text)

        # remove empty paragraphs:
        text = re.sub(r"[\r\n]+# *[\r\n]+", "\n", text)

        # adjust spacing after closing brackets and punctuation:
        fmt = ")»،؛:.!؟\-"
        fmt2 = fmt + "\d\s"
        text = re.sub("([{}]+)([^{}])".format(fmt, fmt2), r"\1 \2", text)
        text = re.sub("\) ([{}])".format(fmt), r")\1", text)

        # restore the formatting of urls:
        def restore_url_formatting(m):
            descr = m.group(1)
            link = m.group(2)
            link = re.sub(" |~~|\r|\n", "", link)
            return "!"+descr+link
        text = re.sub("(?:# )*! (\[[^\]]*\]) *(\([^\)]+\))", restore_url_formatting, text)

        # adjust spacing before opening brackets:
        text = re.sub("(\w)([(«])", r"\1 \2", text)

        # restore formulae that had zwnj in between them:
        d = {
            "صلىاللهعليهوسلم": "صلى الله عليه وسلم",
            "صلىاللهعليهاوسلم": "صلى الله عليها وسلم",
            "صلىاللهعليهموسلم": "صلى الله عليهم وسلم",
            "صلىاللهعليهماوسلم": "صلى الله عليهما وسلم",
            "رحمهالله": "رحمه الله",
            "رحمهاالله": "رحمها الله",
            "رحمهمالله": "رحمهم الله",
            "رحمهماالله": "رحمهما الله",
            "عليهالسلام": "عليه السلام",
            "عليهمالسلام": "عليهم السلام",
            "عليهماالسلام": "عليهما السلام",
            "عليهاالسلام": "عليها السلام",
            "عزوجل": "عز وجل",
            "صلىاللهعليهوآلهوسلم": "صلى الله عليه وآله وسلم",
            "صلىاللهعليهوآله": "صلى الله عليه وآله",
            }
        formula_regex = "|".join(d.keys())
        text = re.sub(formula_regex, lambda m: d.get(m.group(0)), text)

        # remove superfluous new lines before a new paragraph/page number
        text = re.sub("[\r\n]+(# |Page)", r"\n\1", text)

        return text


if __name__ == "__main__":
    folder = r"test/Rafed/"
    conv = RafedHtmlConverter()
    #conv.convert_file(folder+"Rafed_3943.html")
    convert_files_in_folder(folder)
    input("DONE")
    import doctest
    doctest.testmod()
    input("Passed all tests. Continue?")





"""Convert Ghaemiyeh library html to OpenITI mARkdown.

This script subclasses the generic MarkdownConverter class
from the html2md module (based on python-markdownify,
https://github.com/matthewwithanm/python-markdownify),
which uses BeautifulSoup to create a flexible converter.

The subclass in this module, QaemiyehConverter,
adds methods specifically for the conversion of books from
the Qaemiyeh library to OpenITI mARkdown.

There do not seem to be many special cases in Ghaemiyeh texts, 
only footnotes: <div class="calibre1" id="content_note_12_1">1 footnotetext</div>

The easiest way to use this is to simply feed the html (as string)
to the markdownify() function, which will create an instance
of the QaemiyehConverter class and return the converted string.

Examples (doctests):

    Headings: h1 (from superclass html2md.MarkdownConverter)

    >>> import html2md_Qaemiyeh
    >>> h = '<h1>abc</h1>'
    >>> html2md_Qaemiyeh.markdownify(h)
    '\\n\\n### | abc\\n\\n'

    NB: heading style is OpenITI mARkdown style by default,
        but can be set to other styles as well:

    >>> h = '<h1>abc</h1>'
    >>> html2md_Qaemiyeh.markdownify(h, md_style=UNDERLINED)
    '\\n\\nabc\\n===\\n\\n'

    >>> h = '<h1>abc</h1>'
    >>> html2md_Qaemiyeh.markdownify(h, md_style=ATX)
    '\\n\\n# abc\\n\\n'


    Footnote divs:
    
    >>> h = '<div class="calibre1" id="content_note_12_1">1 footnotetext</div>'
    >>> html2md_Qaemiyeh.markdownify(h)
    '\\n\\nFOOTNOTE1 footnotetext\\n\\n'

    NB: FOOTNOTE is a tag that will be used to extract all footnotes
        in a next step.

    Divs without class or with an unsupported class are simply stripped:
    
    >>> h = 'abc\
             <div>def</div>\
             ghi'
    >>> html2md_Qaemiyeh.markdownify(h)
    'abc def ghi'
        
    >>> h = 'abc\
             <div class="unknown_div_class">def</div>\
             ghi'
    >>> html2md_Qaemiyeh.markdownify(h)
    'abc def ghi'

    Spans without class or with an unsupported class are stripped:
    
    >>> h = 'abc <span>def</span> ghi'
    >>> html2md_Qaemiyeh.markdownify(h)
    'abc def ghi'
    
    >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
    >>> html2md_Qaemiyeh.markdownify(h)
    'abc def ghi'


    Links: 

    >>> h = '<a href="a/b/c">abc</a>'
    >>> html2md_Qaemiyeh.markdownify(h)
    '[abc](a/b/c)'

    
    Unordered lists: 

    >>> h = '<ul><li>item1</li><li>item2</li></ul>'
    >>> html2md_Qaemiyeh.markdownify(h)
    '\\n* item1\\n* item2\\n\\n'
    
    Ordered lists:

    >>> h = '<ol><li>item1</li><li>item2</li></ol>'
    >>> html2md_Qaemiyeh.markdownify(h)
    '\\n1. item1\\n2. item2\\n\\n'

    Nested lists:
    
    >>> h = '<ol><li>item1</li><li>item2:<ul><li>item3</li><li>item4</li></ul></li></ol>'
    >>> html2md_Qaemiyeh.markdownify(h)
    '\\n1. item1\\n2. item2:\\n\\n\\t* item3\\n\\t* item4\\n\\t\\n\\n'


    Italics (<i> and <em> tags):

    >>> h = 'abc <em>def</em> ghi'
    >>> html2md_Qaemiyeh.markdownify(h)
    'abc *def* ghi'

    >>> h = 'abc <i>def</i> ghi'
    >>> html2md_Qaemiyeh.markdownify(h)
    'abc *def* ghi'


    Bold (<b> and <strong> tags):

    >>> h = 'abc <b>def</b> ghi'
    >>> html2md_Qaemiyeh.markdownify(h)
    'abc **def** ghi'

    >>> h = 'abc <strong>def</strong> ghi'
    >>> html2md_Qaemiyeh.markdownify(h)
    'abc **def** ghi'


    Tables:
    
    >>> h = '\
    <table>\
      <tr>\
        <th>th1aaa</th><th>th2</th>\
      </tr>\
      <tr>\
        <td>td1</td><td>td2</td>\
      </tr>\
    </table>'
    >>> html2md_Qaemiyeh.markdownify(h)
    '\\n\\n| th1aaa | th2 |\\n| ------ | --- |\\n| td1    | td2 |\\n\\n'

"""
import re


if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(path.dirname(root_folder)))
    sys.path.append(root_folder)

from openiti.new_books.convert.helper import html2md
from openiti.new_books.convert.helper.html2md import *  # import all constants!

import re


class QaemiyehConverter(html2md.MarkdownConverter):
    """Convert Qaemiyeh library html to OpenITI mARkdown."""

    def __init__(self, **options):
        super().__init__(**options)
        self.class_dict = dict()
        #self.class_dict["quran"] = '@QUR@ {text}\n'
        #self.class_dict["hadith"] = '@HAD@ {text}\n'
        #self.class_dict["subtitle"] = 'DELETE_PREVIOUS_BLANKLINES {text}\n\n'


    #------------------------------------------------------------------------
    # new conversion methods 
    # (and adaptations of conversion methods from the super-class)

    def convert_a(self, el, text):
        """Converts html links.

        Overwrites the MarkdownConverter.post_process_md() method.
        Introduces an exception for links between footnote markers
        and footnootes. 

        Example:
            >>> import html2md_Qaemiyeh
            >>> h = '<a href="a/b/c">abc</a>'
            >>> html2md_Qaemiyeh.markdownify(h)
            '[abc](a/b/c)'

            >>> import html2md_Qaemiyeh
            >>> h = 'abc <a href="ftn1">1</a>'
            >>> html2md_Qaemiyeh.markdownify(h)
            'abc [1]'
        """
        href = el.get('href')
        if href:
            if "content_note" in href or "fn" in href:
                # do not include a link for footnotes:
                return text
            else:
                # return a markdown representation of a link: [text](href)
                return super().convert_a(el, text)
        else:
            return super().convert_a(el, text)

    def convert_div(self, el, text):
        """Converts html <div> tags, depending on their class.

        In the MarkdownConverter class, div tags are simply stripped away.
        
        Examples:
        
            # no div class: tags are stripped off
            
            >>> import html2md_Qaemiyeh
            >>> h = 'abc\
                     <div>def</div>\
                     ghi'
            >>> html2md_Qaemiyeh.markdownify(h)
            'abc def ghi'
            
            # unknown div class: tags are stripped off
            
            >>> import html2md_Qaemiyeh
            >>> h = 'abc\
                     <div class="unknown_div_class">def</div>\
                     ghi'
            >>> html2md_Qaemiyeh.markdownify(h)
            'abc def ghi'
            

            # footnote:
            
            >>> h = '<div class="calibre1" id="content_note_24_1>1 footnotetext</div>'
            >>> html2md_Qaemiyeh.markdownify(h)
            '\\n\\nFOOTNOTE1 footnotetext\\n\\n'

        """
        
        try:
            if el["id"].startswith("content_note_"):
                return "\n\nFOOTNOTE" + text + "\n\n"
        except:
            pass
        try:
            el["class"]
            for c in el["class"]:
                if c in self.class_dict:
                    return self.class_dict[c].format(text=text, el=el, c=c) 
        except:
            return text
           
        return text

    def convert_p(self, el, text):
        """Converts html <p> tags, depending on their class attribute.

        Supported span classes should be stored in self.span_dict
        (key: span class (str); value: formatting string)

        Example:
            >>> import html2md_Qaemiyeh
            >>> h = '<p>abc def ghi</p>'
            >>> html2md_Qaemiyeh.markdownify(h)
            '# abc def ghi'
            
            >>> h = 'abc <p class="unknown_span_class">def</p> ghi'
            >>> html2md_Qaemiyeh.markdownify(h)
            'abc \n#def ghi'
        """
        try:
            for c in el["class"]:
                if c in self.class_dict:
                    if text:
                        return self.class_dict[c].format(text=text, el=el, c=c)
                    else:
                        return ''
        except Exception as e:
            #print(e)
            pass
        return "\n# " + text + "\n"


    def convert_span(self, el, text):
        """Converts html <span> tags, depending on their class attribute.

        Supported span classes should be stored in self.span_dict
        (key: span class (str); value: formatting string)

        Example:
            >>> import html2md_Qaemiyeh
            >>> h = 'abc <span>def</span> ghi'
            >>> html2md_Qaemiyeh.markdownify(h)
            'abc def ghi'
            
            >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
            >>> html2md_Qaemiyeh.markdownify(h)
            'abc def ghi'
        """
        try:
            for c in el["class"]:
                if c in self.class_dict:
                    if text:
                        return self.class_dict[c].format(text=text, el=el, c=c)
                    else:
                        return ''
        except Exception as e:
            #print(e)
            pass
        return text



def markdownify(html, **options):
    """Shortcut to the convert method of the QaemiyehConverter class."""
    return QaemiyehConverter(**options).convert(html)


if __name__ == "__main__":
    import doctest
    doctest.testmod()






